//--------------------------------------------------------------------------
// AlgorithmHLDA.java 6.3 01/23/2008
// Created       : Tao Ma
//
// Description   : Describes the HLDA algorithm
// Remarks       : Code converted from matlab
//--------------------------------------------------------------------------


//----------------------
// import java packages
//----------------------
import java.util.*;
import java.awt.*;

/**
 * Implements the Heteroscedastic Linear Discriminant Analysis Algorithm
 */
public class AlgorithmHLDA extends Algorithm
{
    // Public Data Members
    //
    Vector<MyPoint> decision_regions_d;
    Vector<MyPoint> support_vectors_d;
    int output_canvas_d[][];

    // covtype
    int covtype_d;
    // hlda_initype
    int hlda_initype_d;
    
    // declare local Matrix objects
    //
    int ppc;
    int p;
    int N;
    int TN;
    int n;
    int J;
    Matrix HLDA;
    Matrix theta;
    Matrix NJ;
    Matrix Wj;
    Matrix WjB;
    Matrix W;
    Matrix WB;
    Matrix T;
    Matrix TB;
    Matrix x;
    Matrix Tx;
    Matrix xj;
    Matrix xv;
    Matrix Xjb;
    Matrix Xb;
    Matrix classes;
    Matrix Tclasses;
    Matrix options;
    int t0, t1, t2, t3, t4, t5, t6, t7, t8;
    double[] v0, v1, v2, v3, v4, v5, v6, v7, v8;
    Matrix m0, m1, m2, m3, m4, m5, m6, m7, m8;

    
   /**
    * Overrides the initialize() method in the base class. Initializes
    * member data and prepares for execution of first step. This method
    * "resets" the algorithm.
    *
    * @return   Returns true.
    */
    public boolean initialize()
    {
	algo_id = "AlgorithmHLDA";
	
	// Debug 
	//
	System.out.println(algo_id + " initialize()");

	step_count = 3;
	point_means_d      = new Vector<MyPoint>();
	decision_regions_d = new Vector<MyPoint>();
	support_vectors_d  = new Vector<MyPoint>();
	description_d      = new Vector<String>();

	// read covtype from menu configuration
	covtype_d = Classify.main_menu_d.covtype;
	System.out.println("value of covtype (covtype_d): " + covtype_d);
	// read hlda_initype from menu configuration
	hlda_initype_d = Classify.main_menu_d.hlda_initype;
	System.out.println("value of hlda_initype (hlda_initype_d): " + hlda_initype_d);

	// Initialize local Matrix objects
	//
	HLDA  = new Matrix();
	theta  = new Matrix();
	
	// Add the process description for the HLDA algorithm
	//
	if (description_d.size() == 0)
        {
	    String str = new String("   0. Initialize the original data.");
	    description_d.addElement(str);
	    str = new String("   1. Displaying the original data.");
	    description_d.addElement(str);
	    
	    str = new String("   2. Computing the means and covariance.");
	    description_d.addElement(str);
	    
	    str = new String("   3. Computing the decision regions based on the class independent HLDA algorithm.");
	    description_d.addElement(str);
	}
	
	// append message to process box
	//
	pro_box_d.appendMessage("HLDA Analysis:" + "\n");
	
	// set the data points for this algorithm
	set1_d = data_points_d.dset1;
	set2_d = data_points_d.dset2;
	set3_d = data_points_d.dset3;
	set4_d = data_points_d.dset4;

	// Initilize matrices
	//
	NJ = new Matrix();
	Wj = new Matrix();
	WjB = new Matrix();
	WB = new Matrix();
	T = new Matrix();
	TB = new Matrix();
	x = new Matrix();
	Tx = new Matrix();
	xj = new Matrix();
	xv = new Matrix();
	Xjb = new Matrix();
	Xb = new Matrix();
	classes = new Matrix();
	Tclasses = new Matrix();
	options = new Matrix();

	//
	m0 = new Matrix();
	m1 = new Matrix();
	m2 = new Matrix();
	m3 = new Matrix();
	m4 = new Matrix();
	m5 = new Matrix();
	m6 = new Matrix();
	m7 = new Matrix();
	m8 = new Matrix();

	// Size of the matrices
	//

	//
	WB.row = 2;
	WB.col = 2;
	WB.Elem = new double[WB.row][WB.col];
	WB.resetMatrix();
	//
	T.row = 2;
	T.col = 2;
	T.Elem = new double[T.row][T.col];
	T.resetMatrix();
	//
	TB.row = 2;
	TB.col = 2;
	TB.Elem = new double[TB.row][TB.col];
	TB.resetMatrix();

	//
	Xb.row = 2;
	Xb.col = 1;
	Xb.Elem = new double[Xb.row][Xb.col];
	Xb.resetMatrix();

	//
	options.row = 1;
	options.col = 18;
	options.Elem = new double[options.row][options.col];
	options.resetMatrix();

	// n -- the feature dimension
	n = 2;

	// The dimension of the sub-space that contains 
	// discrimination information
        int order = 2;
	p = order;

	// The random problem will correspond to one of 
	// the three cases:
	//    1. All class variances are the same
	//    2. Class variances are diagonalizable by 
	//       the same transformation
	//    3. Class variances are un-equal
	// 
	//int case1 = 3;

	//number of classes
	J = 0;
	if ( set1_d.size()>0 ) {
	    J++;
	    N = set1_d.size();
	    if ( set2_d.size()>0 ) {
		J++;
		N = set1_d.size()+set2_d.size();
		if ( set3_d.size()>0 ) {
		    J++;
		    N = set1_d.size()+set2_d.size()+set3_d.size();
		    if ( set4_d.size()>0 ) {
			J++;
			N = set1_d.size()+set2_d.size()+set3_d.size()+set4_d.size();
		    }
		}
	    }
	}

	//
	NJ.row = 1;
	NJ.col = J;
	NJ.Elem = new double[NJ.row][NJ.col];
	NJ.resetMatrix();
	//
	Wj.row = 2*J;
	Wj.col = 2;
	Wj.Elem = new double[Wj.row][Wj.col];
	Wj.resetMatrix();
	//
	WjB.row = 2*J;
	WjB.col = 2;
	WjB.Elem = new double[WjB.row][WjB.col];
	WjB.resetMatrix();
	//
	Xjb.row = 2;
	Xjb.col = J;
	Xjb.Elem = new double[Xjb.row][Xjb.col];
	Xjb.resetMatrix();

	/**
	 * makedata();
	 */
	classes.row = J+1;
	classes.col = 1;
	classes.Elem = new double[classes.row][classes.col];
	classes.setValue(0, 0, 0);
	if ( set1_d.size()>0 ) {
	    classes.setValue(1, 0, set1_d.size());
	    if ( set2_d.size()>0 ) {
		classes.setValue(2, 0, set1_d.size()+set2_d.size());
		if ( set3_d.size()>0 ) {
		    classes.setValue(3, 0, set1_d.size()+set2_d.size()+set3_d.size());
		    if ( set4_d.size()>0 ) {
			classes.setValue(4, 0, set1_d.size()+set2_d.size()+set3_d.size()+set4_d.size());
		    }
		}
	    }
	}
	Tclasses.copyMatrix(classes);
	//classes.printMatrix();
	//Tclasses.printMatrix();
	
	x.row = N;
	x.col = 2;
	x.Elem = new double[x.row][x.col];
	x.resetMatrix();
	Tx.copyMatrix(x);
	xj.copyMatrix(x);
	xv.copyMatrix(x);
	if ( set1_d.size()>0 ) {
	    for (int i = 0; i < set1_d.size(); i++) {
		MyPoint point = (MyPoint)set1_d.elementAt(i);
		x.setValue(i, 0, point.x);
		x.setValue(i, 1, point.y);
	    }
	    if ( set2_d.size()>0 ) {
		for (int i = 0; i < set2_d.size(); i++) {
		    MyPoint point = (MyPoint)set2_d.elementAt(i);
		    x.setValue((int)classes.getValue(1, 0)+i, 0, point.x);
		    x.setValue((int)classes.getValue(1, 0)+i, 1, point.y);
		}
		if ( set3_d.size()>0 ) {
		    for (int i = 0; i < set3_d.size(); i++) {
			MyPoint point = (MyPoint)set3_d.elementAt(i);
			x.setValue((int)classes.getValue(2, 0)+i, 0, point.x);
			x.setValue((int)classes.getValue(2, 0)+i, 1, point.y);
		    }
		    if ( set4_d.size()>0 ) {
			for (int i = 0; i < set4_d.size(); i++) {
			    MyPoint point = (MyPoint)set4_d.elementAt(i);
			    x.setValue((int)classes.getValue(3, 0)+i, 0, point.x);
			    x.setValue((int)classes.getValue(3, 0)+i, 1, point.y);
			}
		    }
		}
	    }
	}
	//x.printMatrix();

	/**
	 * initialize();
	 */
	//Initialization
	// a = size(Tx);
	// TN = a(1);	
	// 
	N = x.getNumRows();
	n = x.getNumColumns();

	//
	J = classes.getNumRows() - 1;
	for (int j=1; j<=J; j++)
	{
	    NJ.setValue(0, j-1, classes.getValue(j, 0)-classes.getValue(j-1, 0));
	}

	//Xb = mean(x)';
	Matrix Xb_t = new Matrix();
	x.mean(Xb_t);
	Xb_t.transposeMatrix(Xb);
	
	Xjb.resetMatrix();
	for (int j=1; j<=J; j++)
        {
	    //     Xjb(:,j) = mean(x(classes(j)+1:classes(j+1),:))';
	    t0 = (int) classes.getValue(j-1, 0);
	    t1 = (int) classes.getValue(j, 0) - 1;
	    m0.row = t1 - t0 + 1;
	    m0.col = x.getNumColumns();
	    m0.Elem = new double[m0.row][m0.col];
	    x.copyRows(m0, t0, t1-t0+1);
	    m0.mean(m1);
	    m1.transposeMatrix(m2);
	    v0 = new double[Xjb.row];
	    m2.getColumn(v0, v0.length, 0);
	    Xjb.setColumn(v0, v0.length, j-1);
        }
	//Xjb.printMatrix();

	//     xj(classes(j)+1:classes(j+1),1:n) = x(classes(j)+...
	//         1:classes(j+1),:) - ones(NJ(j),1)*Xjb(:,j)';
	//     Wj(n*(j-1)+1:n*j,:) = xj(classes(j)+...
	//         1:classes(j+1),1:n)'*xj(classes(j)+1:classes(j+1),1:n);
	//     WjB(n*(j-1)+1:n*j,:) = Wj(n*(j-1)+1:n*j,:)/NJ(j);
	//
	for (int j = 1; j <= J; j++) {
	    int start = (int)classes.getValue(j - 1, 0) + 1;
	    int end = (int)classes.getValue(j, 0);
	    Matrix tmp_xj = new Matrix();
	    tmp_xj.row = end - start + 1;
	    tmp_xj.col = n;
	    tmp_xj.Elem = new double[tmp_xj.row][tmp_xj.col];
	    tmp_xj.resetMatrix();
	    Matrix tmp_Wj = new Matrix();
	    tmp_Wj.row = n;
	    tmp_Wj.col = n;
	    tmp_Wj.Elem = new double[tmp_Wj.row][tmp_Wj.col];
	    tmp_Wj.resetMatrix();
	    Matrix tmp_mat = new Matrix();
	    tmp_mat.row = n;
	    tmp_mat.col = end - start + 1;
	    tmp_mat.Elem = new double[tmp_mat.row][tmp_mat.col];
	    tmp_mat.resetMatrix();
	    for (int k = start; k <= end; k++) {
		for (int r = 1; r <= n; r ++) {
		    double tmp_diff = x.getValue(k - 1, r - 1)
			- Xjb.getValue(r - 1, j - 1);
		    xj.setValue(k - 1, r - 1, tmp_diff);
		}
	    }
	    xj.copyRows(tmp_xj, start - 1, end - start + 1);
	    tmp_xj.transposeMatrix(tmp_mat);
	    tmp_mat.multMatrix(tmp_xj, tmp_Wj);
	    for (int k = 1; k <= n; k++) {
		for (int r = 1; r <= n; r ++) {
		    Wj.setValue(n * (j - 1) + k - 1, r - 1,
				tmp_Wj.getValue(k - 1, r - 1));
		    WjB.setValue(n * (j - 1) + k - 1, r - 1,
				 (tmp_Wj.getValue(k - 1, r - 1)  /
				   NJ.getValue(0, j - 1)));
		}
	    }
	}
	//xj.printMatrix();
	//Wj.printMatrix();
	//WjB.printMatrix();

	//W = zeros(n,n);
	W = new Matrix();
	W.row = 2;
	W.col = 2;
	W.Elem = new double[W.row][W.col];
	W.resetMatrix();
	for (int j=1; j<=J; j++)
        {
	    t0 = n*(j-1);
	    t1 = n*j - 1;
	    m0.copyMatrix(W);
	    Wj.copyRows(m0, t0, t1-t0+1);
	    W.addMatrix(m0);
	}
	//W.printMatrix();

	// WB = W/N;
	WB.copyMatrix(W);
	WB.scalarMultMatrix((double)1/(double)N);

	//xv = x - ones(N,1)*Xb';
	m0.initMatrixValue(N, 1, 1.0, 0);
	Xb.transposeMatrix(m1);
	m2.copyMatrix(x);
	m0.multMatrix(m1, m2);
	x.subtractMatrix(m2, xv);

	//T = xv'*xv;
	xv.transposeMatrix(m0);
	m0.multMatrix(xv, T);

	//TB = T/N;
	TB.copyMatrix(T);
	TB.scalarMultMatrix((double)1/(double)N);

	//B = T-W;
	Matrix B = new Matrix();
	B.copyMatrix(T);
	T.subtractMatrix(W, B);

	//theta
	theta.row = n;
	theta.col = n;
	theta.Elem = new double[theta.row][theta.col];
	theta.resetMatrix();

	//% This is new data. So the theta have to be computed.
	

	//% These are Numerical optimization options %
	//options = foptions;
	// read directly from matlab
	options.Elem = new double[][] { {0,0.0001,0.0001,0.000001,0,0,0,0,0,0,0,0,0,0,0,1e-8,0.1,0} };

	//% Provide a feedback every options(1) iterations
	//options(1) = 200;
	options.setValue(0, 0, 200.0);
	
	//% Minimum Step size
	//options(2) = 1e-8;
	options.setValue(0, 1, 1e-8);

	//% Minimum improvement at every iteration
	//options(3) = 1e-8;
	options.setValue(0, 2, 1e-8);

	//% Minimum magnitude of the derivative
	//options(4) = 1e-8;
	options.setValue(0, 3, 1e-8);

	//options(6) = 1;
	options.setValue(0, 5, 1);
	
	//options(9) = 0;
	options.setValue(0, 8, 0);

	//options(13) = 0;
	options.setValue(0, 12, 0);

	//% Maximum number of iterations
	//options(14) = 10;%1000000;
	options.setValue(0, 13, 10);

	//options(16) = 1e-11;
	options.setValue(0, 15, 1e-11);
	
	//options(17) = 1e-3;
	options.setValue(0, 16, 1e-3);

	//% Initial guess of step-size
	//options(18) = 0.001;
	options.setValue(0, 17, 0.001);	

	// set the step indexMatrix W;
	//
	step_index_d = 0;
	
	// append message to process box
	//
	pro_box_d.appendMessage((String)description_d.get(step_index_d));
	
	// exit initialize
	//
	return true;
    }

    /**
    * Displays data sets from input box in output box.
    *
    * @return Returns true
    */
    boolean step1()
    {
	// Debug
	//
	System.out.println(algo_id + " step1()");
	
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(20);
	pro_box_d.setProgressCurr(0);

	scaleToFitData();
	
	// append message to process box
	//
	output_panel_d.addOutput(set1_d, Classify.PTYPE_INPUT, 
				 data_points_d.color_dset1);
	output_panel_d.addOutput(set2_d, Classify.PTYPE_INPUT,
				 data_points_d.color_dset2);
	output_panel_d.addOutput(set3_d, Classify.PTYPE_INPUT,
				 data_points_d.color_dset3);
	output_panel_d.addOutput(set4_d, Classify.PTYPE_INPUT, 
				 data_points_d.color_dset4);
	
	// step 1 completed
	//
	pro_box_d.setProgressCurr(20);
	output_panel_d.repaint();
	
	return true;
    }
    
    /**
     * Calculates the within class and between class scatter matrix,
     * transforms the data sets ans displays the mean graphically
     * and numerically
     *
     * @return Returns true
     */
    boolean step2()
    {
	// Debug
	//
	System.out.println(algo_id + " step2()");
	
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(20);
	pro_box_d.setProgressCurr(0);

	// compute means
	computeMeans();



	// Tao
	//
	// different initilization method
	//
	if (hlda_initype_d == 1)
        {
	    //[V,D] = eig(inv(WB)*TB);
	    WB.invertMatrix(m0);
	    m0.multMatrix(TB, m1);
	    Matrix V = new Matrix();
	    V.row = 2;
	    V.col = 2;
	    V.Elem = new double[V.row][V.col];
	    Matrix D = new Matrix();
	    D.row = 2;
	    D.col = 2;
	    D.Elem = new double[D.row][D.col];
	    double eigVal[] = null;
	    double eigVec[] = new double[2];

	    // copy matrix to m2
	    //
	    m2.copyMatrix(m1);

	    eigVal = Eigen.compEigenVal(m2);
	    for (int i = 0; i < 2; i++) {
		Eigen.calcEigVec(m1, eigVal[i], eigVec);
		D.setValue(i, i, eigVal[i]);
		for (int j = 0; j < 2; j++) {
		    // changed by Tao
		    V.Elem[j][i] = eigVec[j];
		}
	    }
	    
	    //[y,I] = sort(real(diag(D)')-1);
	    Matrix y = new Matrix();
	    y.row = 1;
	    y.col = 2;
	    y.Elem = new double[y.row][y.col];
	    Matrix I = new Matrix();
	    I.row = 1;
	    I.col = 2;
	    I.Elem = new double[I.row][I.col];
	    if ( D.getValue(0,0) < D.getValue(1,1) ) {
		y.setValue(0, 0, D.getValue(0,0)-1);
		y.setValue(0, 1, D.getValue(1,1)-1);
		I.setValue(0, 0, 1);
		I.setValue(0, 1, 2);
	    }
	    else {
		y.setValue(0, 1, D.getValue(0,0)-1);
		y.setValue(0, 0, D.getValue(1,1)-1);
		I.setValue(0, 1, 1);
		I.setValue(0, 0, 2);
	    }
	    
	    //[a,b] = size(TB);
	    int a = TB.row;
	    	    
	    //for i = a:-1:1
	    //    theta((a-i+1),:) = V(:,I(i))';
	    //end;
	    for (int i=a; i>=1; i--) {
		t0 = (int) I.getValue(0, i-1);
		v0 = new double[V.row];
		V.getColumn(v0, v0.length, t0-1);
		theta.setRow(v0, v0.length, a-i);
		
	    }
	    
	    //t = det(theta);
	    double t = theta.getValue(0,0)*theta.getValue(1,1) - theta.getValue(0,1)*theta.getValue(1,0);
	    
	    //if t<0
	    //    theta(n,:) = -1*theta(n,:);
	    //end;
	    if (t<0) {
		theta.setValue(n-1, 0, (-1)*theta.getValue(n-1, 0));
		theta.setValue(n-1, 1, (-1)*theta.getValue(n-1, 1));
	    }
	    
	    //theta = theta';
	    m0.copyMatrix(theta);
	    m0.transposeMatrix(theta);
	}
	else
        {
	    // theta = eye();
	    theta.setDiagonal(1.0);
        }

	
	System.out.println("theta = ");
	theta.printMatrix();


	// Tao
	//
	// different covariance matrix type
	//
	if (covtype_d == 1) {
	    // problem1() for equal covariance
	    //
	    //	    Matrix theta1 = new Matrix();
	    //theta1.row = n;
	    //theta1.col = n;
	    //theta1.Elem = new double[theta1.row][theta1.col];
	    //theta1.resetMatrix();

	    //computeHLDA(1, 1, theta1);

	    //System.out.println("theta1 = ");
	    //theta1.printMatrix();


	    // overwrite HLDA
	    //
	    HLDA.copyMatrix(theta);
	}
	else if (covtype_d == 2) {
	    // problem2() for diagonal covariance
	    //
	    Matrix theta2 = new Matrix();
	    theta2.row = n;
	    theta2.col = n;
	    theta2.Elem = new double[theta2.row][theta2.col];
	    theta2.resetMatrix();

	    computeHLDA(2, 2, theta2);

	    //System.out.println("theta2 = ");
	    //theta2.printMatrix();


	    // overwrite HLDA
	    //
	    HLDA.copyMatrix(theta2);
	}
	else if (covtype_d == 3) {
	    // problem3() for not equal covariance
	    //
	    Matrix theta3 = new Matrix();
	    theta3.row = n;
	    theta3.col = n;
	    theta3.Elem = new double[theta3.row][theta3.col];
	    theta3.resetMatrix();

	    computeHLDA(3, 3, theta3);

	    //System.out.println("theta3 = ");
	    //theta3.printMatrix();


	    // overwrite HLDA
	    //
	    HLDA.copyMatrix(theta3);
	}



	System.out.println("HLDA = ");
	HLDA.printMatrix();


	// display Matrces
	displayMatrices();
	
	// display means
	//
	output_panel_d.addOutput(point_means_d, 
				 Classify.PTYPE_OUTPUT_LARGE, Color.black);
	
	// display support vectors
	//
	output_panel_d.addOutput(support_vectors_d, 
				 Classify.PTYPE_INPUT, Color.black );
	
	// display support vectors
	//
	pro_box_d.setProgressCurr(20);
	output_panel_d.repaint();
	
	return true;
    }
    
    
    /**
     * Computes the decision regions and totals the data points in error,
     * as well displays the decision region
     *
     * @return Returns true
     */
    boolean step3()
    {
	// Debug
	//
	System.out.println(algo_id + " step3()");
	
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(20);
	pro_box_d.setProgressCurr(0);

	// compute the decision regions
	//
	computeDecisionRegions();
	
	// compute errors
	//
	computeErrors();

	// display support vectors
	//
	output_panel_d.addOutput( decision_regions_d, 
				  Classify.PTYPE_INPUT, 
				  new Color(255, 200, 0));
	
	output_panel_d.repaint();
	
	return true;
    }


    /**
    * Implementation of the run function from the Runnable interface.
    * Determines what the current step is and calls the appropriate method.
    */
    public void run()
    {
	// Debug
	// 
	System.out.println(algo_id + " run()");

	if (step_index_d == 1)
	{
	    disableControl();
	    step1();
	    enableControl();
	}
	
	else if (step_index_d == 2)
	{
	    disableControl();
	    step2(); 
	    enableControl();
	}
	
	else if (step_index_d == 3)
	{
	    disableControl();
	    step3();
	    pro_box_d.appendMessage("   Algorithm Complete");
	    enableControl(); 
	}
	return;
    }


    /**
     * Display two matrices - covariance matrix and the transformation
     * matrix in the text message window
     */
    public void displayMatrices()
    {
	
	// declare local variables
	//
	double a11 = 0.0;
	double a12 = 0.0;
	double a21 = 0.0;
	double a22 = 0.0;
	
	String text;
	
	// Each class Covariance Matrix
	//
	for (int i=1; i<=J; i++) {
	    a11 = MathUtil.setDecimal(WjB.Elem[(i-1)*2+0][0], 2);
	    a12 = MathUtil.setDecimal(WjB.Elem[(i-1)*2+0][1], 2);
	    a21 = MathUtil.setDecimal(WjB.Elem[(i-1)*2+1][0], 2);
	    a22 = MathUtil.setDecimal(WjB.Elem[(i-1)*2+1][1], 2);
	    text =
		new String(
		       "       Set " + Integer.toString(i) + " covariance matrix:\n"
		       + "         "
		       + a11
		       + "    "
		       + a12
		       + "\n"
		       + "         "
		       + a21
		       + "    "
		       + a22);
	
	    // append message to process box
	    //
	    pro_box_d.appendMessage(text);
	}

	// Overall Covariance Matrix
	//
	a11 = MathUtil.setDecimal(TB.Elem[0][0], 2);
	a12 = MathUtil.setDecimal(TB.Elem[0][1], 2);
	a21 = MathUtil.setDecimal(TB.Elem[1][0], 2);
	a22 = MathUtil.setDecimal(TB.Elem[1][1], 2);
	text =
	    new String(
		       "       Overall covariance matrix:\n"
		       + "         "
		       + a11
		       + "    "
		       + a12
		       + "\n"
		       + "         "
		       + a21
		       + "    "
		       + a22);
	
	// append message to process box
	//
	pro_box_d.appendMessage(text);

	// Transformation matrix
	//
	a11 = MathUtil.setDecimal(HLDA.Elem[0][0], 2);
	a12 = MathUtil.setDecimal(HLDA.Elem[0][1], 2);
	a21 = MathUtil.setDecimal(HLDA.Elem[1][0], 2);
	a22 = MathUtil.setDecimal(HLDA.Elem[1][1], 2);
	
	text =
	    new String(
		       "       Transformation matrix:\n"
		       + "         "
		       + a11
		       + "    "
		       + a12
		       + "\n"
		       + "         "
		       + a21
		       + "    "
		       + a22);
	
	// append message to process box
	//
	pro_box_d.appendMessage(text);
    }


    /**
     * Computes the line of discrimination for class independent HLDA
     */
    public void computeDecisionRegions()
    {
	// Debug
	//
	System.out.println(algo_id + " computeDecisionRegions()");
	
	// local variable 
	//
	MyPoint pt = null;
	double dist = 0.0;
	int associated = 0;
	double smallestSoFar = Double.MAX_VALUE;
	
	int target = 0;
	boolean set1flag = true;
	boolean set2flag = true;
	boolean set3flag = true;
	int counter = 0;
	
	DisplayScale scale = output_panel_d.disp_area_d.getDisplayScale();
	
	double currentX = scale.xmin;
	double currentY = scale.ymin;
	
	// set precision
	//
	int outputWidth = output_panel_d.disp_area_d.getXPrecision();
	int outputHeight = output_panel_d.disp_area_d.getYPrecision();
	
	double incrementY = (scale.ymax - scale.ymin) / outputHeight;
	double incrementX = (scale.xmax - scale.xmin) / outputWidth;
	
	// declare a 2D array to store the class associations
	//
	output_canvas_d = new int[outputWidth][outputHeight];
	
	// loop through each and every point on the pixmap and
	// determine which class each pixel is associated with
	//
	
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(outputWidth);
	
	Matrix vec_S_hat = new Matrix();
	Matrix vec_IS_hat = new Matrix();
	double[] lgs = new double[4];

	vec_S_hat.row = 8;
	vec_S_hat.col = 2;
	vec_S_hat.Elem = new double[vec_S_hat.row][vec_S_hat.col];

	vec_IS_hat.row = 8;
	vec_IS_hat.col = 2;
	vec_IS_hat.Elem = new double[vec_IS_hat.row][vec_IS_hat.col];


	for (int j = 0; j < J; j++) { 
	    //	    m0.copyMatrix(WB);
	    m0.resetMatrix();
	    m0.row = n;
	    m0.col = n;
	    m0.Elem = new double[m0.row][m0.col];

	    if (covtype_d == 1) {
		WB.multMatrix(HLDA, m1);
		HLDA.transposeMatrix(m2);
		m2.multMatrix(m1, m3);
		m3.setValue(0, 1, 0.0);
		m3.setValue(1, 0, 0.0);
	    }
	    else if (covtype_d == 2) {
		//		WjB.scalarMultMatrix(-1.00);
		WjB.copyRows(m0, n * j, n);
		m0.multMatrix(HLDA, m1);
		HLDA.transposeMatrix(m2);
		m2.multMatrix(m1, m3);
		m3.setValue(0, 1, 0.0);
		m3.setValue(1, 0, 0.0);
	    }
	    else {
		//		WjB.scalarMultMatrix(-1.00);
		WjB.copyRows(m0, n * j, n);
		m0.multMatrix(HLDA, m1);
		HLDA.transposeMatrix(m2);
		m2.multMatrix(m1, m3);
	    }
	    
	    lgs[j] = Math.log(m3.getValue(0,0) * m3.getValue(1,1) -
				  m3.getValue(0,1) * m3.getValue(1,0));
	    m3.invertMatrix(m4);
	    
	    double[] tmp_row = new double[n];
	    for (int k = 0; k < n; k++) {
		m3.getRow(tmp_row, n, k);
		vec_S_hat.setRow(tmp_row, n, n * j + k);

		m4.getRow(tmp_row, n, k);
		vec_IS_hat.setRow(tmp_row, n, n * j + k);
	    }
	    //	    vec_S_hat.printMatrix();
	    //vec_IS_hat.printMatrix();
	}

	for (int i = 0; i < outputWidth; i++)
	{
	    currentX += incrementX;
	    currentY = scale.ymin;

	    // set current status
	    //
	    pro_box_d.setProgressCurr(i);
	    
	    for (int j = 0; j < outputHeight; counter++, j++)
	    {
		// declare the current pixel point
		//		    
		currentY += incrementY;
		MyPoint pixel = new MyPoint(currentX, currentY);
		
		// convert the pixel to the time domain
		//
		double X[][] = new double[1][2];
		X[0][0] = pixel.x;
		X[0][1] = pixel.y;
		
		smallestSoFar = Double.MAX_VALUE;
		
		// reset the boolean flags
		//
		set1flag = true;
		set2flag = true;
		set3flag = true;

		// find the closest point from the first class
		//
		for (int k = 0; k < point_means_d.size(); k++)
		{
		    // classify the sample to the first set
		    //
		    if (set1_d.size() > 0 && set1flag)
		    {
			set1flag = false;
			target = 0;
		    }
		    // classify the sample to the second set
		    //
		    else if (set2_d.size() > 0 && set2flag)
		    {
			set2flag = false;
			target = 1;
		    }
		    // classify the sample to the third set
		    //
		    else if (set3_d.size() > 0 && set3flag)
			{
			    set3flag = false;
			    target = 2;
			}
		    // classify the sample to the forth set
		    //
		    else
		    {
			target = 3;
		    }
		    
		    // get the first mean point
		    //
		    pt = (MyPoint)point_means_d.elementAt(k);
		    
		    // convert the mean point to the time domain
		    //
		    double Y[][] = new double[1][2];
		    Y[0][0] = pt.x;
		    Y[0][1] = pt.y;
		    
		    // represent the pixel as a matrix
		    //
		    Matrix A = new Matrix();
		    A.initMatrix(X, 1, 2);
		    
		    // represent the mean point as a matrix
		    //
		    Matrix B = new Matrix();
		    B.initMatrix(Y, 1, 2);
		    
		    // transform the pixel and mean point to
		    // the feature space
		    //
		    Matrix C = new Matrix();
		    Matrix D = new Matrix();
		    
		    A.multMatrix(HLDA, C);
		    B.multMatrix(HLDA, D);

		    Matrix diff = new Matrix();
		    Matrix IS_hat = new Matrix();
		    IS_hat.row = 2;
		    IS_hat.col = 2;
		    IS_hat.Elem = new double[IS_hat.row][IS_hat.col];

		    C.subtractMatrix(D, diff);
		    diff.transposeMatrix(m0);
		    double[] tmp_row = new double[n];
		    for (int row = 0; row < n; row++) {
			vec_IS_hat.getRow(tmp_row, n, k * n + row);
			IS_hat.setRow(tmp_row, n, row);
		    }
		    diff.multMatrix(IS_hat, m1);
		    m1.multMatrix(m0, m2);
		    
		    dist = m2.getValue(0, 0) + lgs[k];
		    //    		    System.out.println(dist);
		    
		    // find the distance between the pixel and
		    // mean point
		    //
		    //		    dist = MathUtil.distance(C.Elem[0][0], C.Elem[0][1], 
		    //					     D.Elem[0][0], D.Elem[0][1]);
		    if (dist < smallestSoFar)
		    {
			associated = target;
			smallestSoFar = dist;
		    }
		}
		
		// put and entry in the output canvas array to
		// indicate which class the current pixel is
		// closest to
		//
		output_canvas_d[i][j] = associated;

		// add a point to the vector of decision
		// region points if the class that the current
		// point is associated with is different for
		// the class what the previous point was
		// associated with i.e., a transition point
		//
		if (j > 0 && i > 0)
		{
		    if (associated != output_canvas_d[i][j - 1] || 
			associated != output_canvas_d[i - 1][j])
		    {
			decision_regions_d.add(pixel);
		    }
		}
	    }
	}
    }   

    /**
     * Counts the data points in each set in error and displays
     * them on the text message window
     */
    public void computeErrors()
    {
	
	// declare local variables
	//
	String text;
	double error;
	int samples = 0;
	int samples1 = 0;
	int samples2 = 0;
	int samples3 = 0;
	int samples4 = 0;
	
	int incorrect = 0;
	int incorrect1 = 0;
	int incorrect2 = 0;
	int incorrect3 = 0;
	int incorrect4 = 0;
	
	DisplayScale scale = output_panel_d.disp_area_d.getDisplayScale();
	
	// set scales
	//
	int outputWidth = output_panel_d.disp_area_d.getXPrecision();
	int outputHeight = output_panel_d.disp_area_d.getYPrecision();
	
	double incrementY = (scale.ymax - scale.ymin) / outputHeight;
	double incrementX = (scale.xmax - scale.xmin) / outputWidth;
	
	// compute the classification error for the first set
	//
	for (int i = 0; i < set1_d.size(); i++)
	{
	    MyPoint point = (MyPoint)set1_d.elementAt(i);
	    samples1++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 0)
		{
		    incorrect1++;
		}
	    }
	}
	
	if (set1_d.size() > 0)
	{
	    
	    error = ((double)incorrect1 / (double)samples1) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 0:\n"
			   + "          Total number of samples: "
			   + samples1
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect1
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the classification error for the second set
	//
	for (int i = 0; i < set2_d.size(); i++)
	{
	    
	    MyPoint point = (MyPoint)set2_d.elementAt(i);
	    samples2++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 1)
		{
		    incorrect2++;
		}
	    }
	}
	
	if (set2_d.size() > 0)
	{
	    
	    error = ((double)incorrect2 / (double)samples2) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 1:\n"
			   + "          Total number of samples: "
			   + samples2
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect2
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the classification error for the third set
	//
	for (int i = 0; i < set3_d.size(); i++)
	{
	    
	    MyPoint point = (MyPoint)set3_d.elementAt(i);
	    samples3++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 2)
		{
		    incorrect3++;
		}
	    }
	}
	
	if (set3_d.size() > 0)
	{
	    error = ((double)incorrect3 / (double)samples3) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 2:\n"
			   + "          Total number of samples: "
			   + samples3
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect3
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the classification error for the forth set
	//
	for (int i = 0; i < set4_d.size(); i++)
	{
	    
	    MyPoint point = (MyPoint)set4_d.elementAt(i);
	    samples4++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 3)
		{
		    incorrect4++;
		}
	    }
	}
	
	if (set4_d.size() > 0)
	{		
	    error = ((double)incorrect4 / (double)samples4) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 3:\n"
			   + "          Total number of samples: "
			   + samples4
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect4
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the overall classification error
	//
	samples = samples1 + samples2 + samples3 + samples4;
	incorrect = incorrect1 + incorrect2 + incorrect3 + incorrect4;
	error = ((double)incorrect / (double)samples) * 100.0;
	
	text =
	    new String(
		       "       Overall results:\n"
		       + "          Total number of samples: "
		       + samples
		       + "\n"
		       + "          Misclassified samples: "
		       + incorrect
		       + "\n"
		       + "          Classification error: "
		       + MathUtil.setDecimal(error, 2)
		       + "%");
	
	pro_box_d.appendMessage(text);
    }



    // computeHLDA
    //
    public void computeHLDA(int fun, int gradfun, Matrix theta1) {
	//tt = fminsd('m1',theta(:),options,'g1');
	Matrix tt = new Matrix();
	fminsd(fun, gradfun, tt);

	//theta = reshape(tt,n,n);
	Matrix theta = new Matrix();
	theta.row = 2;
	theta.col = 2;
	theta.Elem = new double[theta.row][theta.col];
	theta.setValue(0, 0, tt.getValue(0, 0));
	theta.setValue(1, 0, tt.getValue(1, 0));
	theta.setValue(0, 1, tt.getValue(2, 0));
	theta.setValue(1, 1, tt.getValue(3, 0));

	//theta = theta*inv(diag(sqrt(diag(theta*theta'))));
	theta.transposeMatrix(m0);
	theta.multMatrix(m0, m1);
	m2.copyMatrix(m1);
	m2.setValue(0, 1, 0);
	m2.setValue(1, 0, 0);
	m2.sqrtMatrix(m3);
	m3.setValue(0, 1, 0);
	m3.setValue(1, 0, 0);
	m3.invertMatrix(m4);
	theta.multMatrix(m4, m5);
	theta.copyMatrix(m5);

        //theta1 = theta/(det(theta)^(1/n));
	double t0 = theta.getValue(0, 0)*theta.getValue(1, 1) - theta.getValue(0, 1)*theta.getValue(1, 0);
	t0 = Math.sqrt(t0);
	t0 = (double)1/t0;
	theta1.copyMatrix(theta);
	theta1.scalarMultMatrix(t0);


	//debug(t0);
	//theta1.printMatrix();
    }



    // @theta: input tranform matrix
    // @xc:    return tranform matrix
    void fminsd(int fun, int gradfun, Matrix xc) {
	//xc=xi(:)
	xc.row = theta.row * theta.col;
	xc.col = theta.col/2;
	xc.Elem = new double[xc.row][xc.col];
	xc.setValue(0, 0, theta.getValue(0, 0));
	xc.setValue(1, 0, theta.getValue(1, 0));
	xc.setValue(2, 0, theta.getValue(0, 1));
	xc.setValue(3, 0, theta.getValue(1, 1));
	
	//f=eval(evalstr);
	double f = 0;	
	if (fun == 1) {
	    f = fun_m1(xc);
	}
	else if (fun == 2) {
	    f = fun_m2(xc);
	}
	else if (fun == 3) {
	    f = fun_m3(xc);
	}
	else {
	    //
	}

	//n=length(xc);
	
	//nn=sqrt(n);
	
	//g=eval(evalstr2);
	Matrix g = new Matrix();
	if (gradfun == 1) {
	    fun_g1(xc, g);
	}
	else if (gradfun == 2) {
	    fun_g2(xc, g);
	}
	else if (gradfun == 3) {
	    fun_g3(xc, g);
	}
	else {
	    //
	}

	//if OPTIONS(18) >0
	//    step=OPTIONS(18);
	//else
	//    step=0.01;
	//end
	double step = 0;
	if (  options.getValue(0, 17)>0 ) {
	    step = options.getValue(0, 17);
	}
	else {
	    step = 0.01;
	}

	//OLDX=xc;
	Matrix oldx = new Matrix();
	oldx.copyMatrix(xc);

	//OLDF=f;
	double oldf = f;

	//OLDG=g;
	Matrix oldg = new Matrix();
	oldg.copyMatrix(g);

	//nog=norm(OLDG);
	double nog = norm(oldg);

	//ng=nog
	double ng = nog;

		
	//tt=reshape(xc,nn,nn);
	Matrix tt = new Matrix();
	tt.row = 2;
	tt.col = 2;
	tt.Elem = new double[tt.row][tt.col];
	tt.setValue(0, 0, xc.getValue(0, 0));
	tt.setValue(1, 0, xc.getValue(1, 0));
	tt.setValue(0, 1, xc.getValue(2, 0));
	tt.setValue(1, 1, xc.getValue(3, 0));

	//
	//e=0;
	//OPTIONS(10)=1;
	//OPTIONS(11)=1;
	//status=-1;
	//improv=0;
	//disp([f, improv,step, det(reshape(xc,nn,nn))]);
	//cnt=1;
	double e = 0;
	options.setValue(0, 9, 1.0);
	options.setValue(0, 10, 1.0);
	int status = -1;
	double improv = 0;
	int cnt = 1;

	// while loop
	//
	while (status < 2) {
	    if (e < 1) {
	        status = -1;
	    }
	    e = 0;

	    //xc = oldx - step * g;
	    m0.copyMatrix(g);
	    m0.scalarMultMatrix(step);
	    oldx.subtractMatrix(m0, xc);
	    //xc.printMatrix();

	    //f=eval(evalstr);
	    if (fun == 1) {
		f = fun_m1(xc);
	    }
	    else if (fun == 2) {
		f = fun_m2(xc);
	    }
	    else if (fun == 3) {
		f = fun_m3(xc);
	    }
	    else {
		//
	    }
	    //debug(f);

	    options.setValue(0, 10-1, options.getValue(0, 10-1)+1);
	    if (f > oldf) {
		step = step/2.0;
		if ( (step*ng) < options.getValue(0, 2-1) ) {
		    //%disp(['step',sprintf(' %5.5e',step*ng) '<' ...
		    //%sprintf('%5.5e', options(2))]);
		    status = status + 1;
		    e = 1;
		}
		if (options.getValue(0, 10-1) < 80) {
		    status = -1;
		}
		if (options.getValue(0, 14-1) < options.getValue(0, 10-1)) {
		    //%disp(['Max # of iterations exceeded', ...
		    //%sprintf('%7.0f',options(10))]);
		    status = status + 1;
		    e = 1;
		}
	    }
	    else {
		improv = oldf - f;

		//g = eval(evalstr2);
		if (gradfun == 1) {
		    fun_g1(xc, g);
		}
		else if (gradfun == 2) {
		    fun_g2(xc, g);
		}
		else if (gradfun == 3) {
		    fun_g3(xc, g);
		}
		else {
		    //
		}

		options.setValue(0, 11-1, options.getValue(0, 11-1)+1);
		ng = norm(g);

		//% Now a Heuristic formula for step size
		//step = step*0.85*(1+0.6*(sum(oldg.*g)/(nog*ng)));
		step = step*0.85*(1+0.6*(( oldg.getValue(0,0)*g.getValue(0,0)+oldg.getValue(1,0)*g.getValue(1,0)+oldg.getValue(2,0)*g.getValue(2,0)+oldg.getValue(3,0)*g.getValue(3,0) )/(nog*ng)));

		cnt = cnt + 1;
		if (cnt > options.getValue(0, 1-1)) {
		    if (options.getValue(0, 1-1) > 0) {
			//tt = reshape(xc, nn, nn);
			//%disp(sprintf('%d %e %e %e %e',options(10),...
			//%f,improv,step*ng,det(tt)));
			cnt = 1;
		    }
		}
		//oldx = xc;
		oldx.copyMatrix(xc);

		oldf = f;

		//oldg = g;
		oldg.copyMatrix(g);

		nog = ng;
		if ((step * ng) < options.getValue(0, 2-1)) {
		    //%disp(['step',sprintf('%5.5e',step*ng) '<', ...
		    // %sprintf('%5.5e',options(2))]);
		    status = status + 1;
		    e = 1;
		}
		if (improv < options.getValue(0, 3-1)*0.2) {
		    //%disp(['Df',sprintf(' %5.5e',improv) '<', ...
		    //%sprintf(' %5.5e',options(3))]);
		    status = status + 1;
		    e = 1;
		}
		if (ng < options.getValue(0, 4-1)) {
		    //%disp(['G',sprintf(' %5.5e',ng) '<', ...
		    //%sprintf(' %5.5e',options(4))]);
		    status = status + 1;
		    e = 1;
		}
		if (options.getValue(0, 10-1) < 80) {
		    status = -1;
		}
		if (options.getValue(0, 14-1) < options.getValue(0, 10-1)) {
		    //disp(['Max # of iterations exceeded ', ...
		    //	sprintf('%7.0f',options(10))]);
		    status = status + 1;
		    e = 1;
		}
	    }
	}


	//debug(step);
	//xc.printMatrix();

    }


    // To do:
    // implement norm() according to matlab
    //
    double norm(Matrix g) {
	double nog;
	Matrix g_trans = new Matrix();
	Matrix g_sq = new Matrix();
	g_trans.copyMatrix(g);
	g_sq.copyMatrix(g);

	g.transposeMatrix(g_trans);
	g.multMatrix(g_trans, g_sq);

	double eigVal[] = null;
	eigVal = Eigen.compEigenVal(g_sq);
	nog = Math.sqrt(Math.max(eigVal[0], eigVal[1]));

	return nog;
    }

    
    // m1()
    //
    double fun_m1(Matrix tt) {
	//System.out.println("m1() running ... ");

	//theta = reshape(tt,n,n);
	Matrix theta = new Matrix();
	theta.row = 2;
	theta.col = 2;
	theta.Elem = new double[theta.row][theta.col];
	theta.setValue(0, 0, tt.getValue(0, 0));
	theta.setValue(1, 0, tt.getValue(1, 0));
	theta.setValue(0, 1, tt.getValue(2, 0));
	theta.setValue(1, 1, tt.getValue(3, 0));

	//theta_p = theta(:,1:p);
	Matrix theta_p = new Matrix();
	theta_p.row = theta.row;
	theta_p.col = p;
	theta_p.Elem = new double[theta_p.row][theta_p.col];
	theta.copyColumns(theta_p, 0, p);
	double t1;

	if (p < n) {
	    //theta_np = theta(:,p+1:n);
	    Matrix theta_np = new Matrix();
	    theta_np.row = theta.row;
	    theta_np.col = n - p;
	    theta_np.Elem = new double[theta_np.row][theta_np.col];
	    theta.copyColumns(theta_np, p, n-p);

	    //t1 = log(prod(diag(theta_np'*TB*theta_np)));
	    theta_np.transposeMatrix(m0);
	    m0.multMatrix(TB, m1);
	    m1.multMatrix(theta_np, m2);
	    t1 = Math.log(m2.getValue(0, 0));
	}
	else {
	    t1 = 0;
	}

	//t2 = log(prod(diag(theta_p'*WB*theta_p)));
	theta_p.transposeMatrix(m0);
	m0.multMatrix(WB, m1);
	m1.multMatrix(theta_p, m2);
	double t2 = Math.log( m2.getValue(0, 0)*m2.getValue(1, 1) );

	//f = 0.5*(t1+t2)-log(det(theta));
	double f = 0.5*(t1+t2) - Math.log( theta.getValue(0,0)*theta.getValue(1,1)-theta.getValue(0,1)*theta.getValue(1,0) );
	
	//npar = J*p+(n-p)+(n*(n+1)/2);
	double npar = J*p+(n-p)+(n*(n+1)/2);

	//f = f + (npar/2)*(log(N)/N);
	f = f + (npar/2)*(Math.log(N)/N);



	//debug(f);
	//m2.printMatrix();
	
	return f;
    }


    // m2()
    //
    double fun_m2(Matrix tt) {
	//System.out.println("m2() running ... ");

	//theta = reshape(tt,n,n);
	Matrix theta = new Matrix();
	theta.row = 2;
	theta.col = 2;
	theta.Elem = new double[theta.row][theta.col];
	theta.setValue(0, 0, tt.getValue(0, 0));
	theta.setValue(1, 0, tt.getValue(1, 0));
	theta.setValue(0, 1, tt.getValue(2, 0));
	theta.setValue(1, 1, tt.getValue(3, 0));

	//int p = 1;

	//theta_p = theta(:,1:p);
	Matrix theta_p = new Matrix();
	theta_p.row = theta.row;
	theta_p.col = p;
	theta_p.Elem = new double[theta_p.row][theta_p.col];
	theta.copyColumns(theta_p, 0, p);
	double t1 = 0;

	double t2 = 0;
	//for j=1:J
	//    t2 = t2+(NJ(j)/N)*...
	//        log(prod(diag(theta_p'*WjB(n*(j-1)+1:n*j,:)*theta_p)));
	//end;
	for (int j = 1; j <= J; j++) {
	    double prod = 1.0;
	    Matrix tmp_WjB = new Matrix();
	    tmp_WjB.row = n;
	    tmp_WjB.col = n;
	    tmp_WjB.Elem = new double[tmp_WjB.row][tmp_WjB.col];
	    WjB.copyRows(tmp_WjB, n * (j - 1), n);

	    theta_p.transposeMatrix(m0);
	    m0.multMatrix(tmp_WjB, m1);
	    m1.multMatrix(theta_p, m2);
	    for (int i = 0; i < p; i++) {
		prod *= m2.getValue(i, i);
	    }
	    t2 += Math.log(prod) * NJ.getValue(0, j - 1) / N;
	}
	
	//f = 0.5*(t1+t2)-log(det(theta));
	double f = 0.5*(t1+t2) - Math.log( theta.getValue(0,0)*theta.getValue(1,1)-theta.getValue(0,1)*theta.getValue(1,0) );
	
	//npar = 2*J*p+(n-p)+(n*(n-1)/2);
	double npar = 2*J*p+(n-p)+(n*(n-1)/2);

	//f = f + (npar/2)*(log(N)/N);
	f = f + (npar/2)*(Math.log(N)/N);

	return f;
    }


    // m3()
    //
    double fun_m3(Matrix tt) {
	//System.out.println("m3() running ... ");

	//theta = reshape(tt,n,n);
	Matrix theta = new Matrix();
	theta.row = 2;
	theta.col = 2;
	theta.Elem = new double[theta.row][theta.col];
	theta.setValue(0, 0, tt.getValue(0, 0));
	theta.setValue(1, 0, tt.getValue(1, 0));
	theta.setValue(0, 1, tt.getValue(2, 0));
	theta.setValue(1, 1, tt.getValue(3, 0));

	//int p = 1;

	//theta_p = theta(:,1:p);
	Matrix theta_p = new Matrix();
	theta_p.row = theta.row;
	theta_p.col = p;
	theta_p.Elem = new double[theta_p.row][theta_p.col];
	theta.copyColumns(theta_p, 0, p);
	double t1 = 0;

	double t2 = 0;
	//for j=1:J
	//    t2 = t2+(NJ(j)/N)*...
	//        log(prod(diag(theta_p'*WjB(n*(j-1)+1:n*j,:)*theta_p)));
	//end;
	for (int j = 1; j <= J; j++) {
	    double prod = 1.0;
	    Matrix tmp_WjB = new Matrix();
	    tmp_WjB.row = n;
	    tmp_WjB.col = n;
	    tmp_WjB.Elem = new double[tmp_WjB.row][tmp_WjB.col];
	    WjB.copyRows(tmp_WjB, n * (j - 1), n);

	    theta_p.transposeMatrix(m0);
	    m0.multMatrix(tmp_WjB, m1);
	    m1.multMatrix(theta_p, m2);
	    for (int i = 0; i < p; i++) {
		prod *= m2.getValue(i, i);
	    }
	    t2 += Math.log(prod) * NJ.getValue(0, j - 1) / N;
	}
	
	//f = 0.5*(t1+t2)-log(det(theta));
	double f = 0.5*(t1+t2) - Math.log( theta.getValue(0,0)*theta.getValue(1,1)-theta.getValue(0,1)*theta.getValue(1,0) );
	
	//npar = J*p+(n-p)+(J-1)*p*(p+1)/2+(n*(n-1)/2);
	double npar = J*p+(n-p)+(J-1)*p*(p+1)/2+(n*(n-1)/2);

	//f = f + (npar/2)*(log(N)/N);
	f = f + (npar/2)*(Math.log(N)/N);

	return f;
    }


    // g1()
    //
    void fun_g1(Matrix tt, Matrix df) {
	//System.out.println("g1() running ... ");

	//theta = reshape(tt,n,n);
	Matrix theta = new Matrix();
	theta.row = 2;
	theta.col = 2;
	theta.Elem = new double[theta.row][theta.col];
	theta.setValue(0, 0, tt.getValue(0, 0));
	theta.setValue(1, 0, tt.getValue(1, 0));
	theta.setValue(0, 1, tt.getValue(2, 0));
	theta.setValue(1, 1, tt.getValue(3, 0));

	//theta_p = theta(:,1:p);
	Matrix theta_p = new Matrix();
	Matrix theta_np = new Matrix();
	Matrix t1 = new Matrix();
	Matrix t2 = new Matrix();
	Matrix t3 = new Matrix();

	theta_p.row = theta.row;
	theta_p.col = p;
	theta_p.Elem = new double[theta_p.row][theta_p.col];
	theta.copyColumns(theta_p, 0, p);

	if (p < n) {
	    //theta_np = theta(:,p+1:n);

	    theta_np.row = theta.row;
	    theta_np.col = n - p;
	    theta_np.Elem = new double[theta_np.row][theta_np.col];
	    theta.copyColumns(theta_np, p, n-p);

	    //t2 = TB*theta_np*diag(1./diag(theta_np'*TB*theta_np));

	    theta_np.transposeMatrix(m0);
	    m0.multMatrix(TB, m1);
	    m1.multMatrix(theta_np, m2);
	    m2.setValue(0, 0, (double)1/m2.getValue(0, 0));
	    TB.multMatrix(theta_np, m3);
	    m3.multMatrix(m2, m4);
	    t2.copyMatrix(m4);
	    
	}

	//t1 = WB*theta_p*diag(1./diag(theta_p'*WB*theta_p));
	
	theta_p.transposeMatrix(m0);
	m0.multMatrix(WB, m1);
	m1.multMatrix(theta_p, m2);
	m2.setValue(0, 0, (double)1/m2.getValue(0, 0));
	WB.multMatrix(theta_p, m3);
	m3.multMatrix(m2, m4);
	t1.copyMatrix(m4);


	if (p < n) {
	    //t3 = [t1,t2];
	    t3.row = t1.row;
	    t3.col = t1.col + t2.col;
	    t3.Elem = new double[t3.row][t3.col];
	    t3.setValue(0, 0, t1.getValue(0, 0));
	    t3.setValue(1, 0, t1.getValue(1, 0));
	    t3.setValue(0, 1, t2.getValue(0, 0));
	    t3.setValue(1, 1, t2.getValue(1, 0));
	}
	else {
	    t3.copyMatrix(t1);
	}

	//t3 = t3 - inv(theta');
	theta.transposeMatrix(m0);
	m0.invertMatrix(m1);
	t3.subtractMatrix(m1, m2);
	t3.copyMatrix(m2);

	//df = t3(:);
	df.row = t3.row * t3.col;
	df.col = t3.col/2;
	df.Elem = new double[df.row][df.col];
	df.setValue(0, 0, t3.getValue(0, 0));
	df.setValue(1, 0, t3.getValue(1, 0));
	df.setValue(2, 0, t3.getValue(0, 1));
	df.setValue(3, 0, t3.getValue(1, 1));



	//debug(p);
	//df.printMatrix();
    }


    // g2()
    //
    void fun_g2(Matrix tt, Matrix df) {
	//System.out.println("g2() running ... ");

	//theta = reshape(tt,n,n);
	Matrix theta = new Matrix();
	theta.row = 2;
	theta.col = 2;
	theta.Elem = new double[theta.row][theta.col];
	theta.setValue(0, 0, tt.getValue(0, 0));
	theta.setValue(1, 0, tt.getValue(1, 0));
	theta.setValue(0, 1, tt.getValue(2, 0));
	theta.setValue(1, 1, tt.getValue(3, 0));

	//int p = 1;

	//theta_p = theta(:,1:p);
	Matrix theta_p = new Matrix();
		Matrix t2 = new Matrix();
	Matrix t3 = new Matrix();

	theta_p.row = theta.row;
	theta_p.col = p;
	theta_p.Elem = new double[theta_p.row][theta_p.col];
	theta.copyColumns(theta_p, 0, p);


	//t2=0;
	t2.row = p;
	t2.col = p;
	t2.Elem = new double[t2.row][t2.col];
	t2.resetMatrix();

	//for j=1:J
	//    t2 = t2+(NJ(j)/N)*WjB(n*(j-1)+1:n*j,:)*theta_p*...
	//        diag(1./diag(theta_p'*WjB(n*(j-1)+1:n*j,:)*theta_p));
	//end;
	for  (int j = 1; j <= J; j++) {
	    Matrix tmp_WjB = new Matrix();
	    tmp_WjB.row = n;
	    tmp_WjB.col = n;
	    tmp_WjB.Elem = new double[tmp_WjB.row][tmp_WjB.col];
	    WjB.copyRows(tmp_WjB, n * (j - 1), n);

	    theta_p.transposeMatrix(m0);
	    m0.multMatrix(tmp_WjB, m1);
	    m1.multMatrix(theta_p, m2);

	    for (int i = 0; i < m2.row; i++) {
		for (int k =0; k < m2.col; k++) {
		    if (i == k) {
			m2.setValue(i, k, 1 / m2.getValue(i, i));
		    }
		    else
			m2.setValue(i, k, 0.0);
		}
	    }
	    
	    theta_p.multMatrix(m2, m3);
	    tmp_WjB.multMatrix(m3, m4);
	    m4.scalarMultMatrix(NJ.getValue(0, j - 1) / N);
	    t2.addMatrix(m4);
	}

	//t3 = t3 - inv(theta');
	theta.transposeMatrix(m0);
	m0.invertMatrix(m1);
	t3.subtractMatrix(m1, m2);
	t3.copyMatrix(m2);

	//df = t3(:);
	df.row = t3.row * t3.col;
	df.col = t3.col/2;
	df.Elem = new double[df.row][df.col];
	df.setValue(0, 0, t3.getValue(0, 0));
	df.setValue(1, 0, t3.getValue(1, 0));
	df.setValue(2, 0, t3.getValue(0, 1));
	df.setValue(3, 0, t3.getValue(1, 1));	
    }


    // g3()
    //
    void fun_g3(Matrix tt, Matrix df) {
	//System.out.println("g3() running ... ");

	//theta = reshape(tt,n,n);
	Matrix theta = new Matrix();
	theta.row = 2;
	theta.col = 2;
	theta.Elem = new double[theta.row][theta.col];
	theta.setValue(0, 0, tt.getValue(0, 0));
	theta.setValue(1, 0, tt.getValue(1, 0));
	theta.setValue(0, 1, tt.getValue(2, 0));
	theta.setValue(1, 1, tt.getValue(3, 0));

	//theta_p = theta(:,1:p);
	Matrix theta_p = new Matrix();
	Matrix t2 = new Matrix();
	Matrix t3 = new Matrix();

	theta_p.row = theta.row;
	theta_p.col = p;
	theta_p.Elem = new double[theta_p.row][theta_p.col];
	theta.copyColumns(theta_p, 0, p);


	//t2=0;
	t2.row = p;
	t2.col = p;
	t2.Elem = new double[t2.row][t2.col];
	t2.resetMatrix();

	//for j=1:J
	//    t2 = t2+(NJ(j)/N)*WjB(n*(j-1)+1:n*j,:)*theta_p*...
	//        diag(1./diag(theta_p'*WjB(n*(j-1)+1:n*j,:)*theta_p));
	//end;
	for  (int j = 1; j <= J; j++) {
	    Matrix tmp_WjB = new Matrix();
	    tmp_WjB.row = n;
	    tmp_WjB.col = n;
	    tmp_WjB.Elem = new double[tmp_WjB.row][tmp_WjB.col];
	    WjB.copyRows(tmp_WjB, n * (j - 1), n);

	    theta_p.transposeMatrix(m0);
	    m0.multMatrix(tmp_WjB, m1);
	    m1.multMatrix(theta_p, m2);

	    m2.invertMatrix(m3);
	    m2.copyMatrix(m3);

	    theta_p.multMatrix(m2, m3);
	    tmp_WjB.multMatrix(m3, m4);
	    m4.scalarMultMatrix(NJ.getValue(0, j - 1) / N);
	    t2.addMatrix(m4);
	}

	//t3 = t3 - inv(theta');
	theta.transposeMatrix(m0);
	m0.invertMatrix(m1);
	t3.subtractMatrix(m1, m2);
	t3.copyMatrix(m2);

	//df = t3(:);
	df.row = t3.row * t3.col;
	df.col = t3.col/2;
	df.Elem = new double[df.row][df.col];
	df.setValue(0, 0, t3.getValue(0, 0));
	df.setValue(1, 0, t3.getValue(1, 0));
	df.setValue(2, 0, t3.getValue(0, 1));
	df.setValue(3, 0, t3.getValue(1, 1));
    }


    // debug functions to print result
    //
    static void debug(int t0) {
	System.out.println(t0);
    }
    static void debug(double t1) {
	System.out.println(t1);
    }
    static void debug(double[] v0) {
	for (int i=0; i<v0.length; i++) {
	    System.out.println(v0[i]);
	}
    }




}
