//--------------------------------------------------------------------------
// AlgorithmKMeans.java 6.0 03/15/2005
// Created       : Phil Trasatti
//                                   Edited : Sanjay Patil
//                              Last Edited : Ryan Irwin
//
// Description   : Describes the K nearest neighbor algorithms
// Remarks       : Code unchanged since created. Created 07/15/2003
//--------------------------------------------------------------------------


//----------------------
// import java packages
//----------------------
import java.util.*;
import java.awt.*;

/**
 * implements the K nearest neighbor algorithms
 */
public class AlgorithmKMeans extends Algorithm
{

    //-------------------------------------------------------------------
    //
    // Public Data Members
    //
    //-------------------------------------------------------------------
  
    /**
     * Current object index
     */
    int currObject; 

    /**
     * Stores points used in Algorithm
     */
    Vector<MyPoint> guesses;

    /**
     * Stores points used in Algorithm
     */
    Vector<MyPoint> data_pool_d;
    
    /**
     * Stores points of various decision regions computed
     */
    DecisionRegion decision_regions_d;

    /**
     *  The random number generator
     */
    public final static long RAND_SEED = 20022000;

    /**
     * Random number initialized
     */
    Random random = new Random(RAND_SEED);

    /**
     * ID used for message appending. Initialized to "AlgorithmKMeans".
     */
    String algo_id = "AlgorithmKMeans";

    /**
     * Number of iterations.  Initialized to 10.
     */
    int iterations = 10;

    /**
     * Number of guess. Initialized to 4.
     */
    int numguesses = 4;


    //-------------------------------------------------------------------
    //
    // Public methods
    //
    //-------------------------------------------------------------------
  

    /** 
     * Overrides the initialize() method in the base class.  Initializes
     * member data and prepares for execution of first step.  This method
     * "resets" the algorithm.
     *
     * @return  Returns false
     */
    public boolean initialize()
    {
	// Debug
	//
	// System.out.println(algo_id + ": initialize()");
	
	
	scale = output_panel_d.disp_area_d.getDisplayScale();
	data_pool_d = new Vector<MyPoint>();
	decision_regions_d = new DecisionRegion();
	point_means_d = new Vector<MyPoint>();
	// support_vectors_d = new Vector();
	iterations = Classify.main_menu_d.iterations;
	numguesses = Classify.main_menu_d.clusters;
	description_d = new Vector<String>();
	step_count = 3;
	
	currObject = 0;
	
	// add the process description for the LBG algorithm
	//
	if (description_d.size() == 0)
	{
	    String str = new String("   0. Initialize the original data.");
	    description_d.addElement(str);
	    
	    str = new String("   1. Displaying the original data.");
	    description_d.addElement(str);
	    
	    str = new String("   2. Computing the means for each data set.");
	    description_d.addElement(str);
	    
	    str = new String("   3. Stepping through iterations ");
	    description_d.addElement(str);
	    
	    str = new String("Iteration");
	    description_d.addElement(str);		
	}
	
	// append message to process box
	//
	pro_box_d.appendMessage("KMeans Analysis:" + "\n");
	
	// set the data points for this algorithm
	//
	//	set1_d = (Vector)data_points_d.dset1.clone();
	//	set2_d = (Vector)data_points_d.dset2.clone();
	//	set3_d = (Vector)data_points_d.dset3.clone();
	//	set4_d = (Vector)data_points_d.dset4.clone();
	//	
	set1_d = data_points_d.dset1;
	set2_d = data_points_d.dset2;
	set3_d = data_points_d.dset3;
	set4_d = data_points_d.dset4;

	// advance to step 1
	//
	step_index_d = 0;
	
	// append message to process box and scroll
	//
	pro_box_d.appendMessage((String)description_d.get(step_index_d));
	pro_box_d.scrollPane.getVerticalScrollBar().setValue(1000000);

	// exit gracefully
	//
	return false;
    }

    /** 
     * Displays data sets from input box in output box.
     *
     * @return  Returns true
     */
    boolean step1()
    {
	// Debug
	//
	// System.out.println(algo_id + ": step1()");
	
	// set up progress bar
	//
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(1);
	pro_box_d.setProgressCurr(0);
	
	scaleToFitData();
	
	// Display original data
	//
	output_panel_d.addOutput(set1_d, Classify.PTYPE_INPUT, 
				 data_points_d.color_dset1);
	output_panel_d.addOutput(set2_d, Classify.PTYPE_INPUT,
				 data_points_d.color_dset2);
	output_panel_d.addOutput(set3_d, Classify.PTYPE_INPUT,
				 data_points_d.color_dset3);
	output_panel_d.addOutput(set4_d, Classify.PTYPE_INPUT, 
				 data_points_d.color_dset4);
	
	// step 1 completed
	//
	pro_box_d.setProgressCurr(1);
	output_panel_d.repaint();
	
	return true;
    }
    
    /** 
     * Computes the means of each data set and displays the means graphically
     * and numerically
     *
     * @return  Returns true
     */
    boolean step2()
    {
	// Debug
	//
	// System.out.println(algo_id + ": step2()");
	
	computeMeans();
	
	generatePool();
	
	generateMeans(numguesses);
	
	initializeKmeans();
	
	output_panel_d.addOutput(guesses, Classify.PTYPE_OUTPUT_LARGE, 
				 Color.black);	   
	output_panel_d.repaint();
	return true;
    }

    /** 
     * Computes the Decision Regions and the associated errors. 
     *
     * @return  Returns true
     */      
    boolean step3()
    {
	// Debug
	//
	// System.out.println(algo_id + ": step3()");
	
	if (currObject < iterations)
	{
	    pro_box_d.appendMessage("      Iteration " 
				    + (currObject+1) + "\n");	
	    if ((currObject + 1) != iterations)
		step_index_d--;

	    currObject++;
	    
	    decision_regions_d = new DecisionRegion();
		    
	    // classify the data based on the guesses
	    //
	    classify(decision_regions_d);
	    
	    // compute the new means of the classified data
	    //
	    computeMeans(decision_regions_d);

	    // determine the total number of clusters
	    //
	    int numclusters = decision_regions_d.getNumRegions();
	    
	    // determine the classification error of each cluster
	    //
	    double error = 0.0;
	    double total = 0.0;
	    
	    for (int i = 0; i < numclusters; i++) 
	    {
		
		// get the cluster associated with the region
		//
		Vector cluster = decision_regions_d.getRegion(i);
		
		// determine the mean of the current cluster
		//
		MyPoint mean = MathUtil.computeClusterMean(cluster);
		
		// display the mean for the current cluster
		//
		double xval = MathUtil.setDecimal(mean.x, 2);
		double yval = MathUtil.setDecimal(mean.y, 2);
		
		String message = new String("      Mean for cluster " +
					    i + ": " 
					    + xval + ", " + yval);

		pro_box_d.appendMessage(message + "\n");
		
		// determine the covariance matrix for the cluster
		//
		double x[] = new double[cluster.size()];
		double y[] = new double[cluster.size()];
		
		for (int j = 0; j < cluster.size(); j++) 
		{
		    MyPoint covarPoint =  (MyPoint)cluster.elementAt(j);
		    x[j] = covarPoint.x;
		    y[j] = covarPoint.y;
		}
		
		// declare the covariance object
		//
		Covariance cov = new Covariance();
		
		// declare the covariance matrix
		//
		Matrix covar = new Matrix();
		covar.row = covar.col = 2;
		covar.Elem = new double[2][2];
		
		// compute the covariance matrix of the first data set
		//
		covar.Elem = cov.computeCovariance(x, y);
		
		// display the covariance matrix for the cluster
		//
		double c11 = MathUtil.setDecimal(covar.Elem[0][0], 2);
		double c12 = MathUtil.setDecimal(covar.Elem[0][1], 2);
		double c21 = MathUtil.setDecimal(covar.Elem[1][0], 2);
		double c22 = MathUtil.setDecimal(covar.Elem[1][1], 2);
		
		message = new String("      Covariance matrix:\n" +
				       "         " + c11 
				     + "    "      + c12 + "\n" +
				     "         "   + c21 
				     + "    "      + c22);
		
		pro_box_d.appendMessage(message + "\n");
		
		// determine the closest of the original data sets
		//
		int closest = getClosestSet(mean);
		
		// compute the classification error
		//
		total += (double)cluster.size();
		error += (double)displayClusterError(closest, cluster, i);
	    }
	    
	    double err = (error / total) * 100.0;
	    
	    // display the clasification error
	    //
	    String message =  new String("       Overall results:\n" +
					 "          Total number of samples: "
					 + total + "\n" +
					 "          Misclassified samples: "
					 + error + "\n" +
					 "          Classification error: " + 
					 MathUtil.setDecimal(err, 2) + "%");
	    pro_box_d.appendMessage(message + "\n");
	    
	    // update the output canvas
	    //
	}
	
	outputDecisionRegion();
	return true;
    }
    
    /** 
    * Implementation of the run function from the Runnable interface.
    * Determines what the current step is and calls the appropriate method.
    */ 
    public void run()
    {
	// Debug
	//
	// System.out.println(algo_id + ": run()");
	
	if (step_index_d == 1)
	{
	    disableControl();
	    step1();
	    enableControl();
	}
	else if (step_index_d == 2)
	{
	    disableControl();
	    step2(); 
	    enableControl();
	}
	else if (step_index_d == 3)
	{
	    disableControl();
	    step3();
	    enableControl();
	}
	    
	// exit gracefully
	//
	return;
	}

    /**
     * Collects all the data points of all the data sets  
     */
    public void generatePool() 
    {
	// determine the pool of points only once
	//
	if (data_pool_d.size() > 0) 
        {
	    return;
	}

	// add all points from the first data set
	//
	for (int i = 0; i < set1_d.size(); i++) 
	{	    
	    data_pool_d.addElement((MyPoint)set1_d.elementAt(i));
	}

	// add all points from the second data set
	//
	for (int i = 0; i < set2_d.size(); i++) 
	{	    
	    data_pool_d.addElement((MyPoint)set2_d.elementAt(i));
	}

	// add all points from the third data set
	//
	for (int i = 0; i < set3_d.size(); i++) 
	{	    
	    data_pool_d.addElement((MyPoint)set3_d.elementAt(i));
	}

	// add all points from the fourth data set
	//
	for (int i = 0; i < set4_d.size(); i++) 
	{	    
	    data_pool_d.addElement((MyPoint)set4_d.elementAt(i));
	}
    }

    /**
     * Generates random initial guesses (means) for the data set
     *
     * @param numMeans number of mean points
     */
    public void generateMeans(int numMeans) 
    {
	// declare local variables
	//
	double xval = 0.0;
	double yval = 0.0;

	// allocate memory to the guesses vector
	//
	guesses = new Vector<MyPoint>(numMeans, 10);

	// generate the required number of initial guesses
	//
	for (int i=0; i<numMeans; i++) {

	    xval = scale.xmin + (random.nextDouble() * 
				 (scale.xmax-scale.xmin));

	    yval = scale.ymin + (random.nextDouble() * 
				 (scale.ymax-scale.ymin));

	    guesses.addElement(new MyPoint(xval, yval));
	}
    }

    /**
     *  Initializes the kmean array with the original data sets
     */
    public void initializeKmeans() 
    {	
	// add the first data set to the region
	//
	if (set1_d.size() > 0) 
	{
	    decision_regions_d.addRegion(set1_d);
	}	    

	// add the second data set to the region
	//
	if (set2_d.size() > 0) 
	{
	    decision_regions_d.addRegion(set2_d);
	}	    

	// add the third data set to the region
	//
	if (set3_d.size() > 0) 
	{
	    decision_regions_d.addRegion(set3_d);
	}	    

	// add the forth data set to the region
	//
	if (set4_d.size() > 0) 
	{
	    decision_regions_d.addRegion(set4_d);
	}	    

	// add the initial guesses to the region
	//
	decision_regions_d.setGuesses(guesses);

	// add the region to the kmean vector
	//
	// kmeans.addElement(region);
    }

    /** 
     * Classifies the data sets based on the k-means iterative algorithm
     *
     * @param  region - stored data sets from the classification
     * @see    DecisionRegion
     */
    public void classify(DecisionRegion region) 
    {
	// iterate over all points in the current data pool
	//
	for (int i = 0; i < data_pool_d.size(); i++) 
	{
	    
	    // varables to determine the closest point to the mean
	    //
	    double dist = 0.0;
	    int associated = 0;
	    double smallestSoFar = Double.MAX_VALUE;
	    
	    // retrieve a point from the pool
	    //
	    MyPoint point = (MyPoint)data_pool_d.elementAt(i);
	    
	    // iterate over all guesses - means
	    //
	    for (int j = 0; j < guesses.size(); j++) 
	    {
		
		// retrieve one of the guesses
		//
		MyPoint mean = (MyPoint)guesses.elementAt(j);
		
		// determine the distance of the point from the mean
		//
		dist = MathUtil.distance(point.x, point.y, mean.x, mean.y); 
		
		// store the smallest distance
		//
		if(dist < smallestSoFar) 
		{
		    associated = j;
		    smallestSoFar = dist;
		}
	    }
	    
	    // store the point based on the classification
	    //
	    String name = new String("cluster");
	    name = name + associated;
	    region.addPoint(point, name);
	}
    }

    /**
    * Computes the means of the data sets after each iteraion
    * 
    * @param  region - classified data sets
    * @see    DecisionRegion
    */
    public void computeMeans(DecisionRegion region) 
    {
       	// determine the number of classified data sets
	//
	int numsets = region.getNumRegions();

	// remove all previous guesses - mean points
	//
	guesses.removeAllElements();
	
	// iterate over the data sets to determine the new means
	//
	for (int i = 0; i < numsets; i++) 
	{

	    // retrieve the classified region
	    //
	    Vector dataset = (Vector)region.getRegion(i);
	    
	    // determine the mean of the classified region
	    //
	    MyPoint mean = MathUtil.computeClusterMean(dataset);

	    // add the computed mean as one of the new guesses
	    //
	    guesses.addElement(new MyPoint(mean.x, mean.y));
	}

	// add the newly determined guesses to the region
	//
	region.setGuesses(guesses);
    }

    /**
    * Determines the closest data sets to the cluster
    *
    * @param   mean mean point of the cluster
    * 
    * @return  closest data set to the cluster
    * @see     MyPoint
    */
    public int getClosestSet(MyPoint mean) 
    {
	// declare local variables
	//
	double dist1 = 0.0;
	double dist2 = 0.0;
	double dist3 = 0.0;
	double dist4 = 0.0;

	MyPoint mean1;
	MyPoint mean2;
	MyPoint mean3;
	MyPoint mean4;
	int i=0;

	// determine the distance for the cluster to each data set
	//
	if (set1_d.size() > 0)
	{
	    mean1 = (MyPoint)point_means_d.elementAt(i);
	    i++;
	    dist1 = MathUtil.distance(mean.x, mean.y, 
				      mean1.x, mean1.y);
	}
	else
	    dist1 = Float.MAX_VALUE;

	if (set2_d.size() > 0)
	{
	    mean2 = (MyPoint)point_means_d.elementAt(i);
	    i++;
	    dist2 = MathUtil.distance(mean.x, mean.y, 
				      mean2.x, mean2.y);
	}
	else
	    dist2 = Float.MAX_VALUE;

	if (set3_d.size() > 0)
	{
	    mean3 = (MyPoint)point_means_d.elementAt(i);
	    i++;
	    dist3 = MathUtil.distance(mean.x, mean.y, 
				      mean3.x, mean3.y);
	}
	else
	    dist3 = Float.MAX_VALUE;

	if (set4_d.size() > 0)
	{
	    mean4 = (MyPoint)point_means_d.elementAt(i);
	    i++;
	    dist4 = MathUtil.distance(mean.x, mean.y, 
				      mean4.x, mean4.y);
	}
	else
	    dist4 = Float.MAX_VALUE;

	// the first data set is the closest one
	//
	if (dist1 < dist2 && dist1 < dist3 && dist1 < dist4) 
	{
	    return 1;
	}

	// the second data set is the closest one
	//
	if (dist2 < dist1 && dist2 < dist3 && dist2 < dist4) 
	{
	    return 2;
	}

	// the third data set is the closest one
	//
	if (dist3 < dist1 && dist3 < dist2 && dist3 < dist4) 
	{
	    return 3;
	}

	// the forth data set is the closest one
	//
	return 4;
    }

    /**
     * determines the number of points in error, i.e, not classified
     * by finding the distance of the datapoints from the closest of
     * the vector set
     *
     * @param   closest the data specifying which of the dataset is the
     *                  closest.
     * @param   cluster the data points which form a closest cluster
     * @param   id      unused in current implementation
     * @return  error   number of data points which lie out of the 
     *                  classification
     */
    public int displayClusterError(int closest, Vector cluster, int id) 
    {

	// declare local vaiables
	//
	int total = 0;
	int error = 0;
	boolean present = false;

	// the first set is the closest one - determine the error
	//
	if (closest == 1) 
	{
	    for (int i = 0; i < cluster.size(); total++, i++) 
	    {

		// reset the present flag
		//
		present = false;

		// retrieve the next cluster point
		//
		MyPoint cst = (MyPoint)cluster.elementAt(i);

		for (int j = 0; j < set1_d.size(); j++) 
		{

		    // retrieve the next point from the data set
		    //
		    MyPoint pnt = (MyPoint)set1_d.elementAt(j);

		    // check if the point is present
		    //
		    if (cst.x == pnt.x && cst.y == pnt.y) 
		    {
			present = true;
		    }
		}

		// error if the point is not found in the set
		//
		if (!present) 
		{
		    error++;
		}
	    }
	}

	// the second set is the closest one - determine the error
	//
	if (closest == 2) 
	{
	    for (int i = 0; i < cluster.size(); total++, i++) 
	    {

		// reset the present flag
		//
		present = false;

		// retrieve the next cluster point
		//
		MyPoint cst = (MyPoint)cluster.elementAt(i);

		for (int j = 0; j < set2_d.size(); j++) 
		{

		    // retrieve the next point from the data set
		    //
		    MyPoint pnt = (MyPoint)set2_d.elementAt(j);

		    // check if the point is present
		    //
		    if (cst.x == pnt.x && cst.y == pnt.y) 
		    {
			present = true;
		    }
		}

		// error if the point is not found in the set
		//
		if (!present) 
		{
		    error++;
		}
	    }
	}

	// the third set is the closest one - determine the error
	//
	if (closest == 3) 
	{
	    for (int i = 0; i < cluster.size(); total++, i++) 
	    {

		// reset the present flag
		//
		present = false;

		// retrieve the next cluster point
		//
		MyPoint cst = (MyPoint)cluster.elementAt(i);

		for (int j = 0; j < set3_d.size(); j++) 
		{

		    // retrieve the next point from the data set
		    //
		    MyPoint pnt = (MyPoint)set3_d.elementAt(j);

		    // check if the point is present
		    //
		    if (cst.x == pnt.x && cst.y == pnt.y) 
		    {
			present = true;
		    }
		}

		// error if the point is not found in the set
		//
		if (!present) 
		{
		    error++;
		}
	    }
	}

	// the forth set is the closest one - determine the error
	//
	if (closest == 4) 
	{
	    for (int i = 0; i < cluster.size(); total++, i++) 
	    {

		// reset the present flag
		//
		present = false;

		// retrieve the next cluster point
		//
		MyPoint cst = (MyPoint)cluster.elementAt(i);

		for (int j = 0; j < set4_d.size(); j++) 
		{

		    // retrieve the next point from the data set
		    //
		    MyPoint pnt = (MyPoint)set4_d.elementAt(j);

		    // check if the point is present
		    //
		    if (cst.x == pnt.x && cst.y == pnt.y) 
		    {
			present = true;
		    }
		}

		// error if the point is not found in the set
		//
		if (!present) 
		{
		    error++;
		}
	    }
	}

	// return the misclassified samples
	//
	return (int)error;
    }

    /**
    * Computes the k-mean decision region - nearest neighbor algorithm
    *
    * @param  vec vector of initial guesses
    *
    * @return vector of desision region points
    */
    public Vector<MyPoint> getDecisionRegion(Vector<MyPoint> vec) 
    {

	// declare local variables
	//

	double currentX = scale.xmin;
	double currentY = scale.ymin;
	
	MyPoint point;

	int outputHeight = output_panel_d.disp_area_d.getYPrecision();
	int outputWidth = output_panel_d.disp_area_d.getXPrecision();

	double dist = 0.0;
	double cdist = 0.0;
	double smallestSoFar = Double.MAX_VALUE;

	double incrementY = (scale.ymax-scale.ymin) / outputHeight;
	double incrementX = (scale.xmax-scale.xmin) / outputWidth;

	int associated = 0;
	int outputCanvas[][] = new int[outputWidth][outputHeight];

	Vector<MyPoint> decision = new Vector<MyPoint>(100, 10);

	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(outputWidth);
	pro_box_d.setProgressCurr(0);

	for(int i = 0; i < outputWidth; i++) 
	{
	    
	    currentX += incrementX;
	    currentY = scale.ymin;	    	    

	    pro_box_d.setProgressCurr(i);  

	    for(int j = 0; j < outputHeight; j++) 
	    {
		
		currentY += incrementY;	
		
		MyPoint pixel = new MyPoint(currentX, currentY);	                       smallestSoFar = Double.MAX_VALUE;

		// find the closest point from the guesses
		//
		for (int k = 0; k < vec.size(); k++) 
		{

		    double smallest = Double.MAX_VALUE;

		    point = (MyPoint)vec.elementAt(k);
		    dist = MathUtil.distance(pixel.x, pixel.y, 
					     point.x, point.y); 

		    if(dist < smallestSoFar) 
		    {
			
			int l = 0;

			// find the distance between the cluster mean 
			// and the global mean of the first class
			//			
			if(set1_d.size() > 0) 
			{
			    MyPoint mean1 = (MyPoint)
				            point_means_d.elementAt(l);
			    l++;
			    cdist = MathUtil.distance(point.x, point.y, 
						      mean1.x, mean1.y); 
			    if(cdist < smallest) 
			    {			    
				associated = 0;
				smallest = cdist;
			    }
			}

			// find the distance between the cluster mean 
			// and the global mean of the second class
			//			
			if(set2_d.size() > 0) 
			{
			    MyPoint mean2 = (MyPoint)
				            point_means_d.elementAt(l);
			    l++;
			    cdist = MathUtil.distance(point.x, point.y, 
						      mean2.x, mean2.y); 
			    if(cdist < smallest) 
			    {
				associated = 1;
				smallest = cdist;
			    }
			}

			// find the distance between the cluster mean 
			// and the global mean of the third class
			//			
			if(set3_d.size() > 0) 
			{
			    MyPoint mean3 = (MyPoint)
				            point_means_d.elementAt(l);
			    l++;
			    cdist = MathUtil.distance(point.x, point.y,
						      mean3.x, mean3.y); 
			    if(cdist < smallest) 
			    {
				associated = 2;
				smallest = cdist;
			    }
			}

			// find the distance between the cluster mean 
			// and the global mean of the forth class
			//			
			if(set4_d.size() > 0) 
			{
			    MyPoint mean4 = (MyPoint)
				            point_means_d.elementAt(l);
			    l++;
			    cdist = MathUtil.distance(point.x, point.y,
						      mean4.x, mean4.y); 
			    if(cdist < smallest) 
			    {
				associated = 3;
				smallest = cdist;
			    }
			}
			// set the smallest distance so far
			//
			smallestSoFar = dist;
		    }
		}		

		// put and entry in the output canvas array to
		// indicate which class the current pixel is closest to
		//
		outputCanvas[i][j] = associated;

		// add a point to the vector of decision
		// region points if the class that the current
		// point is associated with is different for
		// the class what the previous point was
		// associated with i.e., a transition point
		//
		if(j > 0 && i > 0) {
		    if(associated != outputCanvas[i][j - 1] || 
		       associated != outputCanvas[i - 1][j]) 
		    {
			decision.addElement(pixel);
		    }
		}
	    }
	}

	// return the vector of decision region points
	//
	return decision;
    }

    /**
     * displays the decision region on the output panel
     */
    public void outputDecisionRegion()
    {
	output_panel_d.clear();
	
	output_panel_d.addOutput(set1_d, Classify.PTYPE_INPUT, 
				 data_points_d.color_dset1);
	output_panel_d.addOutput(set2_d, Classify.PTYPE_INPUT,
				 data_points_d.color_dset2);
	output_panel_d.addOutput(set3_d, Classify.PTYPE_INPUT,
				 data_points_d.color_dset3);
	output_panel_d.addOutput(set4_d, Classify.PTYPE_INPUT, 
				 data_points_d.color_dset4);
	
  	output_panel_d.addOutput(decision_regions_d.getGuesses(), 
				 Classify.PTYPE_OUTPUT_LARGE, 
				 Color.black);

	Vector<MyPoint> decision = getDecisionRegion(guesses);

  	output_panel_d.addOutput(decision, 
				 Classify.PTYPE_INPUT, 
				 Color.black);

	output_panel_d.repaint();
    }    
}

