//--------------------------------------------------------------------------
// AlgorithmLBG.java 6.0 03/15/2005
// Created       : Phil Trasatti     Edited : Daniel May
//                                   Edited : Sanjay Patil
//                              Last Edited : Ryan Irwin
//
// Description   : Describes the LBG algorithms
// Remarks       : Code unchanged since created. Created 07/15/2003
//--------------------------------------------------------------------------


//----------------------
// import java packages
//----------------------
import java.util.*;
import java.awt.*;

/**
 * Implements the class dependent Linear Discriminant Algorithm
 */
public class AlgorithmLDA2 extends Algorithm
{
    
    Vector<MyPoint> decision_regions_d;
    Vector<MyPoint> support_vectors_d;
    int output_canvas_d[][];
    
    // declare local matrix objects
    Matrix W;
    Matrix B1;
    Matrix B2;
    Matrix B3;
    Matrix B4;
    Matrix S1;
    Matrix S2;
    Matrix S3;
    Matrix S4;
    Matrix invW;

    Matrix LDA1;
    Matrix LDA2;
    Matrix LDA3;
    Matrix LDA4;

    Matrix CLDA1; // covariance matrix for CLDA1
    Matrix CLDA2; // covariance matrix for CLDA2
    Matrix CLDA3; // covariance matrix for CLDA3
    Matrix CLDA4; // covariance matrix for CLDA4

    Matrix CLDA; // covariance matrix for CLDA

    /**
    * Overrides the initialize() method in the base class. Initializes
    * member data and prepares for execution of first step. This method
    * "resets" the algorithm.
    *
    * @return   Returns true
    */
    public boolean initialize()
    {
	algo_id = "AlgorithmLDA2";
	    
	// Debug
	//
	// System.out.println(algo_id + ": initialize()");
	step_count = 3;
	point_means_d = new Vector<MyPoint>();
	decision_regions_d = new Vector<MyPoint>();
	support_vectors_d = new Vector<MyPoint>();
	description_d = new Vector<String>();
	    
	// create local matrix objects
	//
	W     = new Matrix();
	B1    = new Matrix();
	B2    = new Matrix();
	B3    = new Matrix();
	B4    = new Matrix();
	S1    = new Matrix();
	S2    = new Matrix();
	S3    = new Matrix();
	S4    = new Matrix();
	CLDA  = new Matrix();
	CLDA1 = new Matrix();
	CLDA2 = new Matrix();
	CLDA3 = new Matrix();
	CLDA4 = new Matrix();
	invW  = new Matrix();
	    
	// Add the process description for the LDA algorithm
	//
	if (description_d.size() == 0)
	{
	    String str = new String("   0. Initialize the original data.");
	    description_d.addElement(str);
	    
	    str = new String("   1. Displaying the original data.");
	    description_d.addElement(str);
	    
	    str = new String("   2. Computing the means and covariance.");
	    description_d.addElement(str);
	    
	    str = new String("   3. Computing the decision regions based on the class dependent LDA algorithm.");
	    description_d.addElement(str);
	}
	
	// append message to process box
	//
	pro_box_d.appendMessage("Class Dependent LDA Analysis:" + "\n");
	    
	// set the data points for this algorithm
	//
	//	set1_d = (Vector)data_points_d.dset1.clone();
	//	set2_d = (Vector)data_points_d.dset2.clone();
	//	set3_d = (Vector)data_points_d.dset3.clone();
	//	set4_d = (Vector)data_points_d.dset4.clone();
	//
	set1_d = data_points_d.dset1;
	set2_d = data_points_d.dset2;
	set3_d = data_points_d.dset3;
	set4_d = data_points_d.dset4;
	    
	// set the step index
	//
	step_index_d = 0;
	    
	// append message to process box
	//
	pro_box_d.appendMessage((String)description_d.get(step_index_d));
	    
	// exit gracefully
	//
	return true;
    }
    
    /**
    * Displays data sets from input box in output box.
    *
    * @return Returns true
    */
    boolean step1()
    {
	// Debug
	//
	// System.out.println(algo_id + ": step1()");
	    
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(20);
	pro_box_d.setProgressCurr(0);
	    
	// append message to process box
	//
	output_panel_d.addOutput(set1_d, Classify.PTYPE_INPUT, 
				 data_points_d.color_dset1);
	output_panel_d.addOutput(set2_d, Classify.PTYPE_INPUT,
				 data_points_d.color_dset2);
	output_panel_d.addOutput(set3_d, Classify.PTYPE_INPUT,
				 data_points_d.color_dset3);
	output_panel_d.addOutput(set4_d, Classify.PTYPE_INPUT, 
				 data_points_d.color_dset4);		

	// step 1 completed
	//
	pro_box_d.setProgressCurr(20);
	output_panel_d.repaint();
	return true;
    }

    /**
     * Calculates the within class and between class scatter matrix,
     * transforms the data sets ans displays the mean graphically
     * and numerically
     *
     * @return Returns true
     */
    boolean step2()
    {
	// Debug
	//
	// System.out.println(algo_id + ": step2()");

	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(20);
	pro_box_d.setProgressCurr(0);
		
	computeMeans();

	// determine the within class scatter matrix
	//
	withinClass(W);
		
	// determine the between class scatter matrix
	//
	betweenClass1(B1, B2, B3, B4);
		
	// determine the ratio of the between class scatter matrix
	// to the within class scatter matrix
	//
	W.invertMatrix(invW);
		
	if (set1_d.size() > 0)
	{
	    invW.multMatrix(B1, S1);
	}
	if (set2_d.size() > 0)
	{
	    invW.multMatrix(B2, S2);
	}
	if (set3_d.size() > 0)
	{
	    invW.multMatrix(B3, S3);
	}
	if (set4_d.size() > 0)
	{
	    invW.multMatrix(B4, S4);
	}
		
	// transform the samples from all data sets
	//
	transformLDA1(S1, S2, S3, S4);

	printMatrices();

	// display means
	//
	output_panel_d.addOutput(point_means_d, 
				 Classify.PTYPE_OUTPUT_LARGE, Color.black);
		
	// display support vectors
	//
	output_panel_d.addOutput(support_vectors_d, 
				 Classify.PTYPE_INPUT, Color.cyan);
		
	// display support vectors
	//
	pro_box_d.setProgressCurr(20);
	output_panel_d.repaint();
	return true;
    }

    /**
     * Computes the decision regions and totals the data points in error,
     * as well displays the decision region
     *
     * @return Returns true
     */
    boolean step3()
    {
	// Debug
	//System.out.println(algo_id + ": step3()");

	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(20);
	pro_box_d.setProgressCurr(0);
		
	// compute the decision regisions
	//
	computeDecisionRegions();
		
	// compute errors
	//
	computeErrors();
		
	// display support vectors
	//
	output_panel_d.addOutput( decision_regions_d, 
				  Classify.PTYPE_INPUT, 
				  new Color(255, 200, 0));
		
	output_panel_d.repaint();
		
	return true;
    }

    /**
    * Implementation of the run function from the Runnable interface.
    * Determines what the current step is and calls the appropriate method.
    */
    public void run()
    {
	// Debug
	//
	// System.out.println(algo_id + " run()");
	    
	if (step_index_d == 1)
	{
	    disableControl();
	    step1();
	    enableControl();
	}
	
	else if (step_index_d == 2)
	{
	    disableControl();
	    step2(); 
	    enableControl();
	}
	    
	else if (step_index_d == 3)
	{
	    disableControl();
	    step3();
	    pro_box_d.appendMessage("   Algorithm Complete");
	    enableControl(); 
	}
	return;
    }

    /**
     * Transforms a given set of points to a new space
     * using the class dependent linear discrimination analysis algorithm
     *
     * @param   S1 covariance matrix of the first class
     * @param   S2 covariance matrix of the second class
     * @param   S3 covariance matrix of the third class
     * @param   S4 covariance matrix of the fourth class
     * @see     Matrix
     */
    public void transformLDA1(Matrix S1, Matrix S2, Matrix S3, Matrix S4)
    {
	// declare local variables
	int size = 0;
	int xsize1 = 0;
	int ysize1 = 0;
	int xsize2 = 0;
	int ysize2 = 0;
	int xsize3 = 0;
	int ysize3 = 0;
	int xsize4 = 0;
	int ysize4 = 0;

	double xval1 = 0.0;
	double yval1 = 0.0;
	double xval2 = 0.0;
	double yval2 = 0.0;
	double xval3 = 0.0;
	double yval3 = 0.0;
	double xval4 = 0.0;
	double yval4 = 0.0;

	
	// declare arrays for the eigenvalues
	double eigVal1[] = null;
	double eigVal2[] = null;
	double eigVal3[] = null;
	double eigVal4[] = null;

	// declare an array to store the eigen vectors
	//
	double eigVec[] = new double[2];


	size = set1_d.size();

	xsize1 += size;
	ysize1 += size;

	for (int i = 0; i < size; i++)
	{
	    MyPoint p = (MyPoint)set1_d.elementAt(i);
	    xval1 += p.x;
	    yval1 += p.y;
	}

	// compute the transformation matrix for the first data set
	//
	if (size > 0)
	{
	    // declare matrix objects
	    //
	    Matrix T = new Matrix();
	    Matrix M = new Matrix();
	    Matrix W = new Matrix();
	    
	    // allocate memory for the matrix elements
	    //
	    T.Elem = new double[2][2];
	    M.Elem = new double[2][2];
	    W.Elem = new double[2][2];
	    
	    // initialize the transformation matrix dimensions
	    //
	    W.row = 2;
	    W.col = 2;
	    
	    // reset the matrices
	    //
	    W.resetMatrix();
	    
	    // initialize the matrix needed to compute the eigenvalues
	    //
	    T.initMatrix(S1.Elem, 2, 2);
	    
	    // make a copy of the original matrix
	    //
	    M.copyMatrix(T);
	    
	    // compute the eigen values
	    //
	    eigVal1 = Eigen.compEigenVal(T);
	    
	    // compute the eigen vectors
	    //
	    for (int i = 0; i < 2; i++)
	    {
		Eigen.calcEigVec(M, eigVal1[i], eigVec);
		for (int j = 0; j < 2; j++)
		{
		    W.Elem[j][i] = eigVec[j];
		}
	    }
	    
	    // save the transformation matrix 
	    //
	    LDA1 = W;
	    
	}
	
	size = set2_d.size();
	
	xsize2 += size;
	ysize2 += size;
	
	for (int i = 0; i < size; i++)
	{	    
	    MyPoint p = (MyPoint)set2_d.elementAt(i);
	    xval2 += p.x;
	    yval2 += p.y;
	}

	// compute the transformation matrix for the second data set
	//
	if (size > 0)
	{   
	    // declare matrix objects
	    //
	    Matrix T = new Matrix();
	    Matrix M = new Matrix();
	    Matrix W = new Matrix();
	    
	    // allocate memory for the matrix elements
	    //
	    T.Elem = new double[2][2];
	    M.Elem = new double[2][2];
	    W.Elem = new double[2][2];
	    
	    // initialize the transformation matrix dimensions
	    //
	    W.row = 2;
	    W.col = 2;
		    
	    // reset the matrices
	    //
	    W.resetMatrix();
	    
	    // initialize the matrix needed to compute the eigenvalues
	    //
	    T.initMatrix(S2.Elem, 2, 2);
	    
	    // make a copy of the original matrix
	    //
	    M.copyMatrix(T);
	    
	    // compute the eigen values
	    //
	    eigVal2 = Eigen.compEigenVal(T);
	    
	    // compute the eigen vectors
	    //
	    for (int i = 0; i < 2; i++)
	    {
		Eigen.calcEigVec(M, eigVal2[i], eigVec);
		for (int j = 0; j < 2; j++)
		{
		    W.Elem[j][i] = eigVec[j];
		}
	    }
	    
	    // save the transformation matrix 
	    //
	    LDA2 = W;
	}
	
	size = set3_d.size();

	xsize3 += size;
	ysize3 += size;

	for (int i = 0; i < size; i++)
	{    
	    MyPoint p = (MyPoint)set3_d.elementAt(i);
	    xval3 += p.x;
	    yval3 += p.y;
	}

	// compute the transformation matrix for the third data set
	//
	if (size > 0)
	{    
	    // declare matrix objects
	    //
	    Matrix T = new Matrix();
	    Matrix M = new Matrix();
	    Matrix W = new Matrix();
	    
	    // allocate memory for the matrix elements
	    //
	    T.Elem = new double[2][2];
	    M.Elem = new double[2][2];
	    W.Elem = new double[2][2];
	    
	    // initialize the transformation matrix dimensions
	    //
	    W.row = 2;
	    W.col = 2;
	    
	    // reset the matrices
	    //
	    W.resetMatrix();
	    
	    // initialize the matrix needed to compute the eigenvalues
	    //
	    T.initMatrix(S3.Elem, 2, 2);
	    
	    // make a copy of the original matrix
	    //
	    M.copyMatrix(T);
		    
	    // compute the eigen values
	    //
	    eigVal3 = Eigen.compEigenVal(T);
	    
	    // compute the eigen vectors
	    //
	    for (int i = 0; i < 2; i++)
	    {
		Eigen.calcEigVec(M, eigVal3[i], eigVec);
		for (int j = 0; j < 2; j++)
		{
		    W.Elem[j][i] = eigVec[j];
		}
	    }
	    
	    // save the transformation matrix 
	    //
	    LDA3 = W;
	}
	
	size = set4_d.size();
	
	xsize4 += size;
	ysize4 += size;
	
	for (int i = 0; i < size; i++)
	{    
	    MyPoint p = (MyPoint)set4_d.elementAt(i);
	    xval4 += p.x;
	    yval4 += p.y;
	}
	
	// compute the transformation matrix for the fourth data set
	//
	if (size > 0)
	{
	    
	    // declare matrix objects
	    //
	    Matrix T = new Matrix();
	    Matrix M = new Matrix();
	    Matrix W = new Matrix();
	    
	    // allocate memory for the matrix elements
	    //
	    T.Elem = new double[2][2];
	    M.Elem = new double[2][2];
	    W.Elem = new double[2][2];
	    
	    // initialize the transformation matrix dimensions
	    //
	    W.row = 2;
	    W.col = 2;
	    
	    // reset the matrices
	    //
	    W.resetMatrix();
	    
	    // initialize the matrix needed to compute the eigenvalues
	    //
	    T.initMatrix(S4.Elem, 2, 2);
	    
	    // make a copy of the original matrix
	    //
	    M.copyMatrix(T);
	    
	    // compute the eigen values
	    //
	    eigVal4 = Eigen.compEigenVal(T);
	    
	    // compute the eigen vectors
	    //
	    for (int i = 0; i < 2; i++)
	    {
		Eigen.calcEigVec(M, eigVal4[i], eigVec);
		for (int j = 0; j < 2; j++)
		{
		    W.Elem[j][i] = eigVec[j];
		}
	    }
	    
	    // save the transformation matrix 
	    //
	    LDA4 = W;
	}
	
    }

    /**  
     * Determines the within class scatter matrix
     *
     * @param   M Matrix storing within class scatter matrix
     * @see     Matrix
     */
    public void withinClass(Matrix M)
    {
	// declare local variables
	//
	int size = 0;
	double x[] = null;
	double y[] = null;
	
		
	// declare the covariance object
	//
	Covariance cov = new Covariance();
	
	
	// get the first data set size
	//
	size = set1_d.size();
	
	// initialize arrays to store the samples
	//
	x = new double[size];
	y = new double[size];
	
	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
	{
	    MyPoint p = (MyPoint)set1_d.elementAt(i);
	    x[i] = p.x;
	    y[i] = p.y;
	}
	
	// compute the covariance matrix of the first data set
	//
	CLDA.row = CLDA1.col = 2;
	CLDA1.Elem = new double[2][2];
	CLDA1.resetMatrix();
	
	if (size > 0)
	{
	    CLDA1.Elem = cov.computeCovariance(x, y);
	}
	
	// get the second data set size
	//
	size = set2_d.size();
	
	// initialize arrays to store the samples
	//
	x = new double[size];
	y = new double[size];
	
	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
        {
	    MyPoint p = (MyPoint)set2_d.elementAt(i);
	    x[i] = p.x;
	    y[i] = p.y;
	}
	
	// compute the covariance matrix of the second data set
	//
	CLDA2.row = CLDA2.col = 2;
	CLDA2.Elem = new double[2][2];
	CLDA2.resetMatrix();
	
	if (size > 0)
	{
	    CLDA2.Elem = cov.computeCovariance(x, y);
	}
	
	// get the third data set size
	//
	size = set3_d.size();
	
	// initialize arrays to store the samples
	//
	x = new double[size];
	y = new double[size];
	
	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
	{
	    MyPoint p = (MyPoint)set3_d.elementAt(i);
	    x[i] = p.x;
	    y[i] = p.y;
	}
	
	// compute the covariance matrix of the third data set
	//
	CLDA3.row = CLDA3.col = 2;
	CLDA3.Elem = new double[2][2];
	CLDA3.resetMatrix();
	
	if (size > 0)
	{
	    CLDA3.Elem = cov.computeCovariance(x, y);
	}
	
	// get the fourth data set size
	//
	size = set4_d.size();
	
	// initialize arrays to store the samples
	//
	x = new double[size];
	y = new double[size];
	
	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
	{
	    MyPoint p = (MyPoint)set4_d.elementAt(i);
	    x[i] = p.x;
	    y[i] = p.y;
	}
	
	// compute the covariance matrix of the fourth data set
	//
	CLDA4.row = CLDA4.col = 2;
	CLDA4.Elem = new double[2][2];
	CLDA4.resetMatrix();
	
	if (size > 0)
        {
	    CLDA4.Elem = cov.computeCovariance(x, y);
	}
	
	// compute the within class scatter matrix
	//
	M.row = M.col = 2;
	M.Elem = new double[2][2];
	M.resetMatrix();
	M.addMatrix(CLDA1);
	M.addMatrix(CLDA2);
	M.addMatrix(CLDA3);
	M.addMatrix(CLDA4);
	CLDA = M;
    }
    
	//  // method dispalyMatrices
	//  //
	//  // arguments:
	//  //    Data d: input data point
	//  //
	//  // return   : none
	//  //
	//  // display two matrices
	//  //
	//  public void displayMatrices()
	//  {
	//
	//    // declare local variables
	//    //
	//    double a11 = 0.0;
	//    double a12 = 0.0;
	//    double a21 = 0.0;
	//    double a22 = 0.0;
	//
	//    String text;
	//
	//    a11 = MathUtil.setDecimal(cov_matrix_d.Elem[0][0], 2);
	//    a12 = MathUtil.setDecimal(cov_matrix_d.Elem[0][1], 2);
	//    a21 = MathUtil.setDecimal(cov_matrix_d.Elem[1][0], 2);
	//    a22 = MathUtil.setDecimal(cov_matrix_d.Elem[1][1], 2);
	//
	//    text =
	//      new String(
	//        "      Covariance matrix:\n"
	//          + "         "
	//          + a11
	//          + "    "
	//          + a12
	//          + "\n"
	//          + "         "
	//          + a21
	//          + "    "
	//          + a22);
	//
	//    // append message to process box
	//    //
	//    pro_box_d.appendMessage(text);
	//
	//    a11 = MathUtil.setDecimal(trans_matrix_d.Elem[0][0], 2);
	//    a12 = MathUtil.setDecimal(trans_matrix_d.Elem[0][1], 2);
	//    a21 = MathUtil.setDecimal(trans_matrix_d.Elem[1][0], 2);
	//    a22 = MathUtil.setDecimal(trans_matrix_d.Elem[1][1], 2);
	//
	//    text =
	//      new String(
	//        "      Transformation matrix:\n"
	//          + "         "
	//          + a11
	//          + "    "
	//          + a12
	//          + "\n"
	//          + "         "
	//          + a21
	//          + "    "
	//          + a22);
	//
	//    // append message to process box
	//    //
	//    pro_box_d.appendMessage(text);
	//
	//  }
    
    /**
     * method computes the line of discrimination for class dependent LDA 
     */
    public void computeDecisionRegions()
    {
	// Debug
	//
	// System.out.println(algo_id + ": computeDecisionRegions()");
	
	// Setup Local Variables
	//
	MyPoint pt = null;
	double dist = 0.0;
	int associated = 0;
	double smallestSoFar = Double.MAX_VALUE;
	
	int target = 0;
	boolean set1flag = true;
	boolean set2flag = true;
	boolean set3flag = true;
	int counter = 0;
	
	DisplayScale scale = output_panel_d.disp_area_d.getDisplayScale();
	
	double currentX = scale.xmin;
	double currentY = scale.ymin;
	
	// set precision
	//
	int outputWidth = output_panel_d.disp_area_d.getXPrecision();
	int outputHeight = output_panel_d.disp_area_d.getYPrecision();
	
	double incrementY = (scale.ymax - scale.ymin) / outputHeight;
	double incrementX = (scale.xmax - scale.xmin) / outputWidth;
	
	// declare a 2D array to store the class associations
	//
	output_canvas_d = new int[outputWidth][outputHeight];
	
	// loop through each and every point on the pixmap and
	// determine which class each pixel is associated with
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(outputWidth);
	
	for (int i = 0; i < outputWidth; i++)
	{
	    currentX += incrementX;
	    currentY = scale.ymin;
	    
	    //set current status
	    pro_box_d.setProgressCurr(i);
	    
	    for (int j = 0; j < outputHeight; counter++, j++)
	    {
		
		currentY += incrementY;
		
		// declare the current pixel point
		//
		MyPoint pix = new MyPoint(currentX, currentY);
		
		// convert the pixel to the time domain
		//
		double X[][] = new double[1][2];
		X[0][0] = pix.x;
		X[0][1] = pix.y;
		
		smallestSoFar = Double.MAX_VALUE;
		
		// reset the boolean flags
		//
		set1flag = true;
		set2flag = true;
		set3flag = true;
		
		// find the closest point from the first class
		//
		for (int k = 0; k < point_means_d.size(); k++)
	        {
		    // get the first mean point
		    //
		    pt = (MyPoint)point_means_d.elementAt(k);
		    
		    // convert the mean point to the time domain
		    //
		    double Y[][] = new double[1][2];
		    Y[0][0] = pt.x;
		    Y[0][1] = pt.y;
		    
		    // represent the pixel as a matrix
		    //
		    Matrix A = new Matrix();
		    A.initMatrix(X, 1, 2);
		    
		    // represent the mean point as a matrix
		    //
		    Matrix B = new Matrix();
		    B.initMatrix(Y, 1, 2);
		    
		    // transform the pixel and mean point to
		    // the feature space
		    //
		    Matrix C = new Matrix();
		    Matrix D = new Matrix();
		    
		    // use the transformation matrix of the
		    // first data set
		    //
		    if (set1_d.size() > 0 && set1flag)
		    {
			set1flag = false;
			target = 0;
			A.multMatrix(LDA1, C);
			B.multMatrix(LDA1, D);
		    }
		    
		    // use the transformation matrix of the
		    // second data set
		    //
		    else if (set2_d.size() > 0 && set2flag)
		    {
			set2flag = false;
			target = 1;
			A.multMatrix(LDA2, C);
			B.multMatrix(LDA2, D);
		    }
		    
		    // use the transformation matrix of the
		    // third data set
		    //
		    else if (set3_d.size() > 0 && set3flag)
		    {
			set3flag = false;
			target = 2;
			A.multMatrix(LDA3, C);
			B.multMatrix(LDA3, D);
		    }
		    
		    // use the transformation matrix of the
		    // forth data set
		    //
		    else
		    {
			target = 3;
			A.multMatrix(LDA4, C);
			B.multMatrix(LDA4, D);
		    }
		    
		    // find the distance between the pixel and
		    // mean point
		    //
		    dist = MathUtil.distance(C.Elem[0][0], C.Elem[0][1], 
					     D.Elem[0][0], D.Elem[0][1]);
		    
		    if (dist < smallestSoFar)
		    {
			associated = target;
			smallestSoFar = dist;
		    }
		}
		
		// put and entry in the output canvas array to 
		// indicate which class the current pixel is closest to
		//
		output_canvas_d[i][j] = associated;
		
		// add a point to the vector of decision
		// region points if the class that the current
		// point is associated with is different for
		// the class what the previous point was
		// associated with i.e., a transition point
		//
		if (j > 0 && i > 0)
		{
		    if (associated != output_canvas_d[i][j - 1] || 
			associated != output_canvas_d[i - 1][j])
		    {
			decision_regions_d.add(pix);
		    }
		}
	    }
	}
    }
    
    /**
     * Determines the between class scatter matrix for
     * the class dependent linear discrimination algorithm
     * 
     * @param   M1 between class scatter matrix for the first class
     * @param   M2 between class scatter matrix for the second class
     * @param   M3 between class scatter matrix for the third class
     * @param   M4 between class scatter matrix for the fourth class
     */
    public void betweenClass1(Matrix M1, Matrix M2, Matrix M3, Matrix M4)
    {
	
	// declare local variables
	//
		
	double xmean1 = 0.0;
	double ymean1 = 0.0;
	double xmean2 = 0.0;
	double ymean2 = 0.0;
	double xmean3 = 0.0;
	double ymean3 = 0.0;
	double xmean4 = 0.0;
	double ymean4 = 0.0;
	double xout1 = 0.0;
	double yout1 = 0.0;
	double xout2 = 0.0;
	double yout2 = 0.0;
	double xout3 = 0.0;
	double yout3 = 0.0;
	double xout4 = 0.0;
	double yout4 = 0.0;
	

	// declare the initial random variables
	//
	double transpose[][] = new double[2][1];
	double mean[][] = new double[1][2];
	
	
	// initialize the between class matrices
	//
	if (set1_d.size() > 0)
       	{
	    M1.row = M1.col = 2;
	    M1.Elem = new double[2][2];
	    M1.resetMatrix();
	}
       	if (set2_d.size() > 0)
       	{
	    M2.row = M2.col = 2;
	    M2.Elem = new double[2][2];
	    M2.resetMatrix();
	}
	if (set3_d.size() > 0)
       	{
	    M3.row = M3.col = 2;
	    M3.Elem = new double[2][2];
	    M3.resetMatrix();
	}
	if (set4_d.size() > 0)
       	{
	    M4.row = M4.col = 2;
	    M4.Elem = new double[2][2];
	    M4.resetMatrix();
	}

	int j = 0;
	
	// obtain the means of each individual class
	//
	if (set1_d.size() > 0)
       	{
	    MyPoint p = (MyPoint)point_means_d.elementAt(j);
	    j++;
	    xmean1 = p.x;
	    ymean1 = p.y;
	}
	
	if (set2_d.size() > 0)
	{
	    MyPoint p = (MyPoint)point_means_d.elementAt(j);
	    j++;
	    xmean2 = p.x;
	    ymean2 = p.y;
	}

       	if (set3_d.size() > 0)
       	{
	    MyPoint p = (MyPoint)point_means_d.elementAt(j);
	    j++;
	    xmean3 = p.x;
	    ymean3 = p.y;
	}
	
	if (set4_d.size() > 0)
	{
	    MyPoint p = (MyPoint)point_means_d.elementAt(j);
	    j++;
	    xmean4 = p.x;
	    ymean4 = p.y;
	}
	
	// compute the mean of all points outside the first set
	//
	if ((set2_d.size() > 0) || (set3_d.size() > 0) || (set4_d.size() > 0))
       	{
	    int samples = 0;
	    
	    for (int i = 0; i < set2_d.size(); samples++, i++)
	    {
		MyPoint p2 = (MyPoint)set2_d.elementAt(i);
		xout1 += p2.x;
		yout1 += p2.y;
	    }
	    for (int i = 0; i < set3_d.size(); samples++, i++)
	    {
		MyPoint p2 = (MyPoint)set3_d.elementAt(i);
		xout1 += p2.x;
		yout1 += p2.y;
	    }
	    for (int i = 0; i < set4_d.size(); samples++, i++)
	    {
		MyPoint p2 = (MyPoint)set4_d.elementAt(i);
		xout1 += p2.x;
		yout1 += p2.y;
	    }
	    xout1 /= samples;
	    yout1 /= samples;
	}
	
	// compute the mean of all points outside the second set
	//
	if ((set1_d.size() > 0) || (set3_d.size() > 0) || (set4_d.size() > 0))
        {
	    
	    int samples = 0;
	    
	    for (int i = 0; i < set1_d.size(); samples++, i++)
	    {			
		MyPoint p2 = (MyPoint)set1_d.elementAt(i);
		xout2 += p2.x;
		yout2 += p2.y;
	    }
	    for (int i = 0; i < set3_d.size(); samples++, i++)
            {
		MyPoint p2 = (MyPoint)set3_d.elementAt(i);
		xout2 += p2.x;
		yout2 += p2.y;
	    }
	    for (int i = 0; i < set4_d.size(); samples++, i++)
	    {
		MyPoint p2 = (MyPoint)set4_d.elementAt(i);
		xout2 += p2.x;
		yout2 += p2.y;
	    }
	    xout2 /= samples;
	    yout2 /= samples;
	}
	
	// compute the mean of all points outside the third set
	//
	if ((set1_d.size() > 0) || (set2_d.size() > 0) || (set4_d.size() > 0))
       	{
	    int samples = 0;
	    
	    for (int i = 0; i < set1_d.size(); samples++, i++)
	    {
		MyPoint p2 = (MyPoint)set1_d.elementAt(i);
		xout3 += p2.x;
		yout3 += p2.y;
	    }
	    for (int i = 0; i < set2_d.size(); samples++, i++)
	    {
		MyPoint p2 = (MyPoint)set2_d.elementAt(i);
		xout3 += p2.x;
		yout3 += p2.y;
	    }
	    for (int i = 0; i < set4_d.size(); samples++, i++)
	    {
		MyPoint p2 = (MyPoint)set4_d.elementAt(i);
		xout3 += p2.x;
		yout3 += p2.y;
	    }
	    xout3 /= samples;
	    yout3 /= samples;
	}

	// compute the mean of all points outside the forth set
	//
	if ((set1_d.size() > 0) || (set2_d.size() > 0) || (set3_d.size() > 0))
       	{   
	    int samples = 0;
	    
	    for (int i = 0; i < set1_d.size(); samples++, i++)
	    {
		
		MyPoint p2 = (MyPoint)set1_d.elementAt(i);
		xout4 += p2.x;
		yout4 += p2.y;
		
	    }
	    for (int i = 0; i < set2_d.size(); samples++, i++)
	    {
		MyPoint p2 = (MyPoint)set2_d.elementAt(i);
		xout4 += p2.x;
		yout4 += p2.y;
		
	    }
	    for (int i = 0; i < set3_d.size(); samples++, i++)
	    {
		MyPoint p2 = (MyPoint)set3_d.elementAt(i);
		xout4 += p2.x;
		yout4 += p2.y;
	    }
	    xout4 /= samples;
	    yout4 /= samples;
	}
	
	// compute the between class scatter of the first set
	//
	if (set1_d.size() > 0)
	{   
	    Matrix S = new Matrix();
	    Matrix M = new Matrix();
	    Matrix T = new Matrix();
	    
	    mean[0][0] = xmean1 - xout1;
	    mean[0][1] = ymean1 - yout1;
	    M.initMatrix(mean, 1, 2);
	    
	    transpose[0][0] = xmean1 - xout1;
	    transpose[1][0] = ymean1 - yout1;
	    T.initMatrix(transpose, 2, 1);
	    
	    T.multMatrix(M, S);
	    M1.addMatrix(S);
	}
	
	// compute the between class scatter contribution of the second set
	//
	if (set2_d.size() > 0)
       	{   
	    Matrix S = new Matrix();
	    Matrix M = new Matrix();
	    Matrix T = new Matrix();
	    
	    mean[0][0] = xmean2 - xout2;
	    mean[0][1] = ymean2 - yout2;
	    M.initMatrix(mean, 1, 2);
	    
	    transpose[0][0] = xmean2 - xout2;
	    transpose[1][0] = ymean2 - yout2;
	    T.initMatrix(transpose, 2, 1);
	    
	    T.multMatrix(M, S);
	    M2.addMatrix(S);
	}
	
	// compute the between class scatter contribution of the third set
	//
	if (set3_d.size() > 0)
	{   
	    Matrix S = new Matrix();
	    Matrix M = new Matrix();
	    Matrix T = new Matrix();
	    
	    mean[0][0] = xmean3 - xout3;
	    mean[0][1] = ymean3 - yout3;
	    M.initMatrix(mean, 1, 2);
	    
	    transpose[0][0] = xmean3 - xout3;
	    transpose[1][0] = ymean3 - yout3;
	    T.initMatrix(transpose, 2, 1);
	    
	    T.multMatrix(M, S);
	    M3.addMatrix(S);
	}
	
	// compute the between class scatter contribution of the fourth set
	//
	if (set4_d.size() > 0)
	{	    
	    Matrix S = new Matrix();
	    Matrix M = new Matrix();
	    Matrix T = new Matrix();
	    
	    mean[0][0] = xmean4 - xout4;
	    mean[0][1] = ymean4 - yout4;
	    M.initMatrix(mean, 1, 2);
	    
	    transpose[0][0] = xmean4 - xout4;
	    transpose[1][0] = ymean4 - yout4;
	    T.initMatrix(transpose, 2, 1);
	    
	    T.multMatrix(M, S);
	    M4.addMatrix(S);
	}
    }
    

    /**
     * Counts the data points in each set in error and displays
     * them on the text message window
     */
    public void computeErrors()
    {
	// declare local variables
	//
	String text;
	double error;
	int samples = 0;
	int samples1 = 0;
	int samples2 = 0;
	int samples3 = 0;
	int samples4 = 0;
	
	int incorrect = 0;
	int incorrect1 = 0;
	int incorrect2 = 0;
	int incorrect3 = 0;
	int incorrect4 = 0;
	
	DisplayScale scale = output_panel_d.disp_area_d.getDisplayScale(); 

	// set scales
	//
	int outputWidth = output_panel_d.disp_area_d.getXPrecision();
	int outputHeight = output_panel_d.disp_area_d.getYPrecision();
	
	double incrementY = (scale.ymax - scale.ymin) / outputHeight;
	double incrementX = (scale.xmax - scale.xmin) / outputWidth;
	
	// compute the classification error for the first set
	//
	for (int i = 0; i < set1_d.size(); i++)
	{
	    
	    MyPoint point = (MyPoint)set1_d.elementAt(i);
	    samples1++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 0)
		{
		    incorrect1++;
		}
	    }
	}
	
	if (set1_d.size() > 0)
	{
	    error = ((double)incorrect1 / (double)samples1) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 0:\n"
			   + "          Total number of samples: "
			   + samples1
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect1
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the classification error for the second set
	//
	for (int i = 0; i < set2_d.size(); i++)
	{
	    
	    MyPoint point = (MyPoint)set2_d.elementAt(i);
	    samples2++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 1)
		{
		    incorrect2++;
		}
	    }
	}
	
	if (set2_d.size() > 0)
	{
	    
	    error = ((double)incorrect2 / (double)samples2) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 1:\n"
			   + "          Total number of samples: "
			   + samples2
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect2
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the classification error for the third set
	//
	for (int i = 0; i < set3_d.size(); i++)
	{  
	    MyPoint point = (MyPoint)set3_d.elementAt(i);
	    samples3++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 2)
		{
		    incorrect3++;
		}
	    }
	}
	
	if (set3_d.size() > 0)
	{   
	    error = ((double)incorrect3 / (double)samples3) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 2:\n"
			   + "          Total number of samples: "
			   + samples3
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect3
			   + "\n"
			       + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the classification error for the forth set
	//
	for (int i = 0; i < set4_d.size(); i++)
	{   
	    MyPoint point = (MyPoint)set4_d.elementAt(i);
	    samples4++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 3)
		{
		    incorrect4++;
		}
	    }
	}
	
	if (set4_d.size() > 0)
	{   
	    error = ((double)incorrect4 / (double)samples4) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 3:\n"
			   + "          Total number of samples: "
			   + samples4
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect4
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the overall classification error
	//
	samples = samples1 + samples2 + samples3 + samples4;
	incorrect = incorrect1 + incorrect2 + incorrect3 + incorrect4;
	error = ((double)incorrect / (double)samples) * 100.0;
	
	text =
	    new String(
		       "       Overall results:\n"
		       + "          Total number of samples: "
		       + samples
		       + "\n"
		       + "          Misclassified samples: "
		       + incorrect
		       + "\n"
		       + "          Classification error: "
		       + MathUtil.setDecimal(error, 2)
		       + "%");
	
	pro_box_d.appendMessage(text);
    }

    /**
     * Display two matrices - covariance matrix and the transformation
     * matrix in the text message window
     */
    public void printMatrices()
    {
	double a11, a12, a21, a22;
	String text;

	if (set1_d.size() > 0)
	{
	    pro_box_d.appendMessage("      Set 1");
	    
	    a11 = MathUtil.setDecimal(CLDA1.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(CLDA1.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(CLDA1.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(CLDA1.Elem[1][1], 2);
	    
	    text = new String("      Covariance matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	    
	    a11 = MathUtil.setDecimal(LDA1.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(LDA1.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(LDA1.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(LDA1.Elem[1][1], 2);
	    
	    text = new String("      Transformation matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	}

	if (set2_d.size() > 0)
	{
	    pro_box_d.appendMessage("      Set 2");
	    
	    a11 = MathUtil.setDecimal(CLDA2.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(CLDA2.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(CLDA2.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(CLDA2.Elem[1][1], 2);
	    
	    text = new String("      Covariance matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	    
	    a11 = MathUtil.setDecimal(LDA2.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(LDA2.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(LDA2.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(LDA2.Elem[1][1], 2);
	    
	    text = new String("      Transformation matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	}

	if (set3_d.size() > 0)
	{
	    pro_box_d.appendMessage("      Set 3");
	    
	    a11 = MathUtil.setDecimal(CLDA3.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(CLDA3.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(CLDA3.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(CLDA3.Elem[1][1], 2);
	    
	    text = new String("      Covariance matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	    
	    a11 = MathUtil.setDecimal(LDA3.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(LDA3.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(LDA3.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(LDA3.Elem[1][1], 2);
	    
	    text = new String("      Transformation matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	}

	if (set4_d.size() > 0)
	{
	    
	    pro_box_d.appendMessage("      Set 4");
	    
	    a11 = MathUtil.setDecimal(CLDA4.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(CLDA4.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(CLDA4.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(CLDA4.Elem[1][1], 2);
	    
	    text = new String("      Covariance matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	    
	    a11 = MathUtil.setDecimal(LDA4.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(LDA4.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(LDA4.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(LDA4.Elem[1][1], 2);
	    
	    text = new String("      Transformation matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	}
    }
}
