/**
 * AlgorithmLP.java v6.0 Last Edited by: Ryan Irwin 05/23/2005
 * Nishant Aggarwal, Jun-Won Suh Created: 12/15/04
 *
 * Description: Linear Prediction algorithm. Determines the estimate of 
 * the data points based on the points given by the user. 
 * The appraoch is data interpolation [ based on cubic interpolation ],
 * Autocorrelation coefficients and Linear Predictor Coefficients [based
 * on Levinson Durbin Algorithm]
 */


//----------------------
// import java packages
//----------------------
import java.util.*;
import java.awt.*;

 
public class AlgorithmLP extends Algorithm
{

    //-----------------------------------------------------------------
    //
    // static data members
    //
    //-----------------------------------------------------------------
    int lporder;
    int iporder;
    int scale;

    //-----------------------------------------------------------------
    //
    // primitive data members
    //
    //-----------------------------------------------------------------
    int output_canvas_d[][];   

    //-----------------------------------------------------------------
    //
    // instance data members
    //
    //-----------------------------------------------------------------
    String algo_id = "AlgorithmLP"; 
    Vector<MyPoint> support_vectors_d;
    Vector<MyPoint> decision_regions_d;

    // for original data
    //
    Vector<MyPoint> set1_d = new Vector<MyPoint>(40, 20);
    Vector<MyPoint> set2_d = new Vector<MyPoint>(40, 20);
    Vector<MyPoint> set3_d = new Vector<MyPoint>(40, 20);
    Vector<MyPoint> set4_d = new Vector<MyPoint>(40, 20);

    // for interpolation function
    //
    Vector<MyPoint> iset1 = new Vector<MyPoint>(40, 20); 
    Vector<MyPoint> iset2 = new Vector<MyPoint>(40, 20); 
    Vector<MyPoint> iset3 = new Vector<MyPoint>(40, 20); 
    Vector<MyPoint> iset4 = new Vector<MyPoint>(40, 20); 

    /**
     * Vector for mean-subtracted [zero-mean] data meaning the original data
     * is interpolated and the average y-value of the set is zero
     */
    Vector<MyPoint> mset1 = new Vector<MyPoint>(40, 20); 
    Vector<MyPoint> mset2 = new Vector<MyPoint>(40, 20); 
    Vector<MyPoint> mset3 = new Vector<MyPoint>(40, 20); 
    Vector<MyPoint> mset4 = new Vector<MyPoint>(40, 20); 

    /**
     * For display purpose meaning it contains original data with points
     * in between
     */
    Vector<MyPoint> display_set1 = new Vector<MyPoint>(40, 20); 
    Vector<MyPoint> display_set2 = new Vector<MyPoint>(40, 20); 
    Vector<MyPoint> display_set3 = new Vector<MyPoint>(40, 20); 
    Vector<MyPoint> display_set4 = new Vector<MyPoint>(40, 20); 

    // to store autoCorrelation coefficient
   //
    double auto_co_1[];
    double auto_co_2[];
    double auto_co_3[];
    double auto_co_4[];

    // to store the average values for each class
    //
    double average1;
    double average2;
    double average3;
    double average4;

    // to store LP coefficient
    //
    double final_lpc_1[]; 
    double final_lpc_2[];
    double final_lpc_3[];
    double final_lpc_4[];

    // to store residual energy 
    //
    double estimate_err_1 = 0;	
    double estimate_err_2 = 0;	
    double estimate_err_3 = 0;
    double estimate_err_4 = 0; 

    // to store actual error energy
    //
    double actual_err_1 = 0;
    double actual_err_2 = 0;
    double actual_err_3 = 0;
    double actual_err_4 = 0;


    // to store the reflection coefficients
    //
    double ref_coeff_1[];
    double ref_coeff_2[];
    double ref_coeff_3[]; 
    double ref_coeff_4[];

    // to store the final results
    //
    Vector<MyPoint> y_estimate1 = new Vector<MyPoint>(40, 20); 
    Vector<MyPoint> y_estimate2 = new Vector<MyPoint>(40, 20); 
    Vector<MyPoint> y_estimate3 = new Vector<MyPoint>(40, 20); 
    Vector<MyPoint> y_estimate4 = new Vector<MyPoint>(40, 20); 

    //---------------------------------------------------------------
    //
    // class methods
    //
    //---------------------------------------------------------------


    /**
     * 
     * Implements the initialize() method in the base class. Initializes
     * member data and prepares for execution of first step. This method
     * "resets" the algorithm.
     *
     * @return   returns true if sets of data are valid
     *
     */
    public boolean initialize()
    {
    	point_means_d           = new Vector<MyPoint>();
	decision_regions_d      = new Vector<MyPoint>();
	support_vectors_d       = new Vector<MyPoint>();
	description_d           = new Vector<String>();

	step_count  = 3;
	lporder    = Classify.main_menu_d.lporder;
	iporder    = Classify.main_menu_d.iporder;
	scale      = lporder * iporder;

	// Add the process description for the LP algorithm
	//
	if (description_d.size() == 0)
	{
	    String str = new String("   0.  Checking for the Data Validity.");
	    description_d.addElement(str);
	    
	    str = new String("   1.  Displaying the Input Data Points.");
	    description_d.addElement(str);
	    
	    str = new String("   2.  Computing LP Parameters.");
	    description_d.addElement(str);
	    
	    str = new String("   3.  Displaying the Linear Predicted Signal");
	    description_d.addElement(str);

	}

	// set the data points for this algorithm
	//
	//	set1_d = (Vector)data_points_d.dset1.clone();
	//	set2_d = (Vector)data_points_d.dset2.clone();
	//	set3_d = (Vector)data_points_d.dset3.clone();
	//	set4_d = (Vector)data_points_d.dset4.clone();
	//
	set1_d = data_points_d.dset1;
	set2_d = data_points_d.dset2;
	set3_d = data_points_d.dset3;
	set4_d = data_points_d.dset4;	

	int max1 = set2_d.size();
	if (set1_d.size() > set2_d.size())
	    max1 = set1_d.size();

	int max2 = set4_d.size();
	if (set3_d.size() > set4_d.size())
	    max2 = set3_d.size();
	else
	    max2 = set4_d.size();

	if (max2 > max1)
	    max1 = max2;

	if (max1 > scale)
	    scale = max1;

	step_index_d = 0;

	if((checkdata_LP(set1_d) == true) && (checkdata_LP(set2_d) == true) 
	   && (checkdata_LP(set3_d) == true) && (checkdata_LP(set4_d) == true))
	{
	    pro_box_d.appendMessage((String)description_d.get(step_index_d));
	    pro_box_d.appendMessage("         Data points valid");
	    return true;	
	}
	else
        {
	    pro_box_d.appendMessage("\n" + "Invalid data");
	    pro_box_d.appendMessage("Clear and enter valid data" + "\n");
	    return false;
	}
    }

    /**
    * 
    * Validates the class entered by user for Linear Prediction  
    *
    * @param    lp
    * @return   true if data is Vector lp is valid. It is invalid
    *           if the size of lp is 1 or any element is larger than the
    *           the previous element
    * 
    */
    public boolean checkdata_LP(Vector lp)
    {
	if ( lp.size() == 1) 
	    return false;
	else
	   for(int i = 0; i <= lp.size() - 1 ; i++ )
	       for(int j = i + 1; j <= lp.size() - 1 ; j++ )
		   // amplitudes should be in time progression
		   //
		   if ( (((MyPoint)lp.elementAt(i)).x == 
			 ((MyPoint)lp.elementAt(j)).x) || 
			(((MyPoint)lp.elementAt(i)).x > 
			 ((MyPoint)lp.elementAt(j)).x) )
		       return false;

	return true;
    }

    /**
     *
     * Implementation of the run function from the Runnable interface.
     * Determines what the current step is and calls the appropriate method.
     *
     */
    public void run()
    {
	if (step_index_d == 1)
	{
	    disableControl();
	    step1();
	    enableControl();
	}
	if (step_index_d == 2)
	{
	    disableControl();
	    step2();
	    enableControl(); 
	}
	if (step_index_d == 3)
	{
	    disableControl();
	    step3(); 
	    enableControl(); 
       }
	
	// exit gracefully
	//
	return;
    }
    
    /**
     *
     * Displays data sets from input box in output box, clears the data
     * points from the Vector fro input data, zero-mean data,
     * does interpolation, and then finds mean of the interpolated data
     * interpolates the zero-mean data, displays the result
     *
     * @return   True
     *
     */
    boolean step1()
    {
	// set up progress bar
	//
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(1);
	pro_box_d.setProgressCurr(0);


      	// Display original data
	// 
       if(set1_d.size() > 0 )  {

	   display_set1.removeAllElements();
	   iset1.removeAllElements();
	   mset1.removeAllElements();

	   interpol(set1_d, display_set1);
	   average1 = mean(display_set1, mset1);
	   interpol(mset1, iset1);

	   output_panel_d.addOutput(set1_d, (Classify.PTYPE_INPUT * 1), 
				    data_points_d.color_dset1);
       }

       if(set2_d.size() > 0 )  {

	   display_set2.removeAllElements();
	   iset2.removeAllElements();  
	   mset2.removeAllElements();

	   interpol(set2_d, display_set2);
	   average2 = mean(display_set2, mset2);
	   interpol(mset2, iset2);   
	  
	   output_panel_d.addOutput(set2_d, (Classify.PTYPE_INPUT * 1), 
				    data_points_d.color_dset2);
       }
       
       if(set3_d.size() > 0 )   {

	   display_set3.removeAllElements();
	   iset3.removeAllElements();
	   mset3.removeAllElements();

	   interpol(set3_d, display_set3);
	   average3 = mean(display_set3, mset3);
	   interpol(mset3, iset3);

	   output_panel_d.addOutput(set3_d, (Classify.PTYPE_INPUT * 1), 
				    data_points_d.color_dset3);
       }
       
       if(set4_d.size() > 0 )
       {	
	   display_set4.removeAllElements();
	   iset4.removeAllElements();
	   mset4.removeAllElements();

	   interpol(set4_d, display_set4);
	   average4 = mean(display_set4, mset4);
	   interpol(mset4, iset4);
   
	   output_panel_d.addOutput(set4_d, (Classify.PTYPE_INPUT * 1), 
				    data_points_d.color_dset4);
       }

	pro_box_d.setProgressCurr(1);
	output_panel_d.repaint();
	return true;
    }
 
    /**
     *
     * Calculates the interpolated points for the data inputs   
     *
     * @param  v  input data points
     * @param  iset  interpolated data points
     *
     */
    public void interpol(Vector<MyPoint> v , Vector<MyPoint> iset)
    {	
	double sample;
	double d ;
	MyPoint r = new MyPoint();	
	
	double[] y2 = new double[v.size()];  
	double[] y  = new double[v.size()];
	double[] x  = new double[v.size()];
       
       for (int k = 0; k < v.size(); k++)
       {
	   x[k] = ((MyPoint)v.elementAt(k)).x;
	   y[k] = ((MyPoint)v.elementAt(k)).y;
       }

       // Reference:
       // "Numerical Recipe in C", Cambridge University Press, pp 113-116
       
       // total time interval of the data points calculated
       //
       d = ((MyPoint)v.lastElement()).x - ((MyPoint)v.firstElement()).x ; 
       
       // time step calculation
       //
       sample = d / scale;
       spline(x, y, y2, v.size());
       
       // iset holds new data points - interpolated values
       // First value added 
       //
       iset.addElement(new MyPoint((MyPoint)v.firstElement()));
       
       // Increment to next Time value
       //
      	r.x = sample + ((MyPoint)v.firstElement()).x;

	while (r.x <=  ((MyPoint)v.lastElement()).x)
	{
	    for(int j = 1; j < v.size(); j++)
	    {
		// the New Interpolated amplitude will be calculated
		// for the two points within which the time intervals 
		// lies, Hence, the condition check
		//
		if((r.x <= ((MyPoint)v.elementAt(j)).x ) && 
		   (r.x >=  ((MyPoint)v.elementAt(j - 1)).x)) 
		{
		    r.y = 0 ;
		    splint ((MyPoint)v.elementAt(j - 1), 
			    (MyPoint)v.elementAt(j), r, y2, (j - 1));

		    // New value appended to the array
		    //
		    iset.addElement(new MyPoint(r));
		    break; 
		}
	    }	    
   	    r.x = r.x + sample;
	}
	// sometimes the last element of the user entered 
	// values is missed out hence the condition to 
	// cross-check and put it in
	//
	if (((MyPoint)iset.lastElement()).x != ((MyPoint)v.lastElement()).x )
	    iset.addElement( new MyPoint((MyPoint)v.lastElement()));
    }

    /**
     *
     * Actually interpolates the points
     *
     * @param x    array containing the x coordinates of datapoints
     * @param y    array containing the y coordinates of datapoints
     * @param y2   array containing the interpolated y coordinates
     * @param size the size of the array to be interpolated
     * 
     */
    public void spline (double[] x, double[] y, double[] y2, int size )
    { 
	double[] u = new double[size];
	double p; 
	double sig; 
	
	y2[0] = 0;
	u[0] = 0;
	 
	for( int i = 1; i < size - 1; i++ )
	{ 
	    sig = ( x[i] - x[i - 1] ) / ( x[i + 1] - x[i - 1] ) ;
	    p = sig * y2[i - 1] + 2 ;
	    y2[i] = ( sig - 1) / p ;
	    u[i] = ( y[i + 1] - y[i] ) / ( x[i + 1] - x[i] ) ;
	    u[i] = ( 6 * u[i] / ( x[i + 1] - x[i - 1] ) 
		     - sig * u[i - 1] ) / p ;
	}
	y2[size - 1] = 0 ;
	
	// This is the back substitution loop of the tridiagonal algorithm
	//
	for(int i = size - 2; i >= 0; i-- )     
	    y2[i] = y2[i] * y2[i + 1] + u[i] ;
    }

    /**
     * Interpolates for a point between the two known points
     * using Cubic Interpolation
     *
     * @param   u1 start point for the interpolation
     * @param   u2 end point for the interpolation
     * @param   r  returning point, basically the interpolated point
     * @param   y2 array used for reassigning of r
     * @param   i  the sample number
     *
     */
    public void splint ( MyPoint u1, MyPoint u2, MyPoint r, 
			 double[] y2, int i )
    {
	double h;
	double a;
	double b;
 
	h = u2.x - u1.x ;
	a = ( u2.x - r.x ) / h ;
	b = ( r.x - u1.x ) / h ;
	r.y = a * u1.y + b * u2.y + ((a * a * a - a) * y2[i] 
	                             + ( b * b * b - b ) * y2[i + 1] ) 
	                                                 * ( h * h ) / 6 ; 
    }


    /**
     *
     * Calculates the mean and the zero-mean data points  
     *
     * @param    v orginal datapoints
     * @param    mv zero mean datapoints 
     *
     * @return   The average in a double
     * 
     */
    public double mean(Vector<MyPoint> v, Vector<MyPoint> mv)
    {
        double average = 0;
	MyPoint p      = new MyPoint();

       for(int i = 0; i < v.size(); i++ )
	   average = average + ((MyPoint)v.elementAt(i)).y;
       average = average / v.size();
       
       for(int i = 0; i < v.size(); i++ )
       {
	   p.y = ((MyPoint)v.elementAt(i)).y - average;
	   p.x =  ((MyPoint)v.elementAt(i)).x; 
	   mv.addElement(new MyPoint(p));
       }        
       return average ;
    }  
   
    /**
    * Computes the Auto Correlation and  Linear Coefficients and 
    * displays the coefficients.
    *
    * @return    True
    *
    */
    boolean step2()
    {
	autoCorrelation();
	lpcCoefficient();
	final_estimate();

	// need a subroutine here to calculate the actual error:
	// 1. Get first two points on the estimated waveform.
	// 2. Check the point given by the user for the x-coordinate,
	//    if the x-coordinate of the user point is same as 
	//    either of two points [from one], 
	// 2a. YES. note the difference in the y-coordinate, this is the error.
	// 2b. Error Energy is square of this term.
	// 2c. NO. check if the user point is outside these two points
	// 2c.1. Yes, pick the next point in the estimated waveform, 
	//       and repeat from step 2 again.
	// 2c.2. No, note the x-coordinate of the user point, 
	//       find the y-coordinate lying on the straight line between
	//       the two estimated points.
	// 2c.2a. Find the difference in the y-coordinates calculate in
	//        step 2c.2 and the user given. This is the error.
	// 2c.3 The Error Energy is the square of this term.
	// 3. Continue with step 2 
	//    till you have covered all the user defined points.
	if(set1_d.size() > 0)
	{
	    actual_err_1 = (double) 0;
	    actual_err_1 = actual_error(y_estimate1, set1_d); 
	}
	if(set2_d.size() > 0)
	{
	    actual_err_2 = (double) 0;
	    actual_err_2 = actual_error(y_estimate2, set2_d); 
	}    
	if(set3_d.size() > 0)
	{
	    actual_err_3 = (double) 0;
	     actual_err_3 = actual_error(y_estimate3, set3_d); 
	}
	if(set4_d.size() > 0)
	{
	    actual_err_4 = (double) 0;
	    actual_err_4 = actual_error(y_estimate4, set4_d); 
	}
	
	step2_display();
	
	return true;
    }    

    /**
    *
    * Computes the autocorrelation coeffient from the data sets   
    *
    */
    public void autoCorrelation()
    {
	if (iset1.size() > 0 )
	{
	    auto_co_1 = new double[lporder + 1];
	    autocorrelate(iset1, auto_co_1);
	}	

	if (iset2.size() > 0)
        {
	    auto_co_2 = new double[lporder + 1];
	    autocorrelate(iset2, auto_co_2);
	}	

	if (iset3.size() > 0)
        {
	    auto_co_3 = new double[lporder + 1];
	    autocorrelate(iset3, auto_co_3);
	}	

	if (iset4.size() > 0)
        {
	    auto_co_4 = new double[lporder + 1];
	    autocorrelate(iset4, auto_co_4);
	}	
    }

    /**
     *
     * Actaully computes the autocorrelation coefficients
     *
     * @param  v Vector of datapoints
     * @param  autoCoeff_co array of autocorrelation coefficients
     *
     */
    public void autocorrelate ( Vector v, double[] autoCoeff_co) 
    {
	int size;
	size                 = v.size();
      	double[] amplitude_y = new double[size];

      	// to store the amplitude for calculation autocoefficient
      	//
    	for (int i = 0; i < size; i++)
	    amplitude_y[i] = ((MyPoint)v.elementAt(i)).y;

      	for (int j = 0; j <= lporder; j++)
      	{
      	    autoCoeff_co[j] = 0;
      	    if ( j < size)
      		for (int k = 0; k < size - j; k++)
		    autoCoeff_co[j] = autoCoeff_co[j] 
			            + amplitude_y[k] * amplitude_y[k + j];
	}

   	// normalize the autocorrelation cofficient
      	//
     	for (int i = lporder; i >= 0; i--) 
      	    autoCoeff_co[i] = autoCoeff_co[i] / autoCoeff_co[0];	  
    }

    /**
     *
     * Computes the Linear Prediction coefficient from the data sets   
     *
     */
    public void lpcCoefficient()
    {
	
	// Rabiner, Schafer, "Digital Processing of Speech signals", 
	// Bell Laboratories, Inc. 1978, 
	// section 8.3.2 Durbin's recursive solution for 
	// the autocorrelation equations pp 411-412
	
	if (set1_d.size() > 0)
	{
	    final_lpc_1 = new double [ lporder + 1];
	    ref_coeff_1 = new double [ lporder + 1];
	    estimate_err_1 = calculate_lpc( auto_co_1, 
					    final_lpc_1, ref_coeff_1);
	}

	if (set2_d.size() > 0)
	{
	    final_lpc_2 = new double [ lporder + 1];
	    ref_coeff_2 = new double [ lporder + 1];
	    estimate_err_2 = calculate_lpc( auto_co_2, 
					    final_lpc_2, ref_coeff_2);
	}

	if (set3_d.size() > 0)
	{
	    final_lpc_3 = new double [ lporder + 1];
	    ref_coeff_3 = new double [ lporder + 1];
	    estimate_err_3 = calculate_lpc( auto_co_3, 
					    final_lpc_3, ref_coeff_3);
	}

	if (set4_d.size() > 0)
	{
	    final_lpc_4 = new double [ lporder + 1];
	    ref_coeff_4 = new double [ lporder + 1];
	    estimate_err_4 = calculate_lpc( auto_co_4, 
					    final_lpc_4, ref_coeff_4);
	}
    }

    /**
     *
     * Actually calculate the LP coefficient and the Residual Error 
     * Energy, and Reflection Coefficients
     *
     * @param  auto_coeff array of auto correlation coefficients
     * @param  lpc        array of linear prediction coefficients
     * @param  rc_reg     array of reflection coefficients
     * 
     * @return residual error energy in a double
     *
     */
    public double calculate_lpc (double[] auto_coeff, 
				 double[] lpc, double[] rc_reg)
    {
	int j_bckwd       = 0;
	long middle_index = 0;

	// Reference:
	// J.D. Markel and A.H. Gray, "Linear Prediction of Speech,"
	// Springer-Verlag Berlin Heidelberg, New York, USA, pp. 219,
	// 1980.
 
	// initialization
	//
	lpc[0] = 1;
	double err_energy = auto_coeff[0];

	// if the energy of the signal is zero we set all the predictor
	// coefficients to zero and exit
	//
	if (err_energy == (double)0)
	{
	    for (int i = 0; i < lporder + 1 ; i++)
		lpc[i] = 0;
	}
	else
        {
	    // do the first step manually
	    //
	    double sum = 0;
	    double tmp = 0;
	    
	    sum         = -auto_coeff[1] / err_energy;
	    rc_reg[1]   = sum;
	    lpc[1]      = rc_reg[1];
	    tmp         = 1 - rc_reg[1] * rc_reg[1];
	    err_energy *= tmp;
	    // recursion
	    //
	    for (int i = 2; i <= lporder; i++)
	    {
		sum = 0;
		for (int j = 1; j < i; j++)
		    sum += lpc[j] * auto_coeff[i - j];
 
		rc_reg[i]    = -(auto_coeff[i] + sum) / err_energy;	    
		lpc[i]       = rc_reg[i];
		j_bckwd      = i - 1;
		middle_index = i / 2;
		for (int j = 1; j <= middle_index; j++)
		{
		    sum          = lpc[j_bckwd] + rc_reg[i] * lpc[j];
		    lpc[j]       = lpc[j] + rc_reg[i] * lpc[i - j];
		    lpc[j_bckwd] = sum;
		    j_bckwd--;
		}
		// compute new error
		//
		tmp = 1.0 - rc_reg[i] * rc_reg[i];
		err_energy *= tmp;
	    }
	}
	return err_energy;
    }

    /**
     *
     * Calculates the estimated points for the data inputs
     *
     */
    public void final_estimate()
    {	
	if(iset1.size() > 0)
	    estimate (iset1, y_estimate1, average1, final_lpc_1);
	
	if(iset2.size() > 0)	    
	    estimate (iset2, y_estimate2, average2, final_lpc_2);
	
	if(iset3.size() > 0)
	    estimate (iset3, y_estimate3, average3, final_lpc_3);
	
	if(iset4.size() > 0)
	    estimate (iset4, y_estimate4, average4, final_lpc_4);
    } 
    
    /**
     *
     * Estimates the amplitude based on the LP coeficients.
     *
     * @param    iset         interpolated data points
     * @param    y_estimate   predicted final signal data points
     * @param    avg          mean of the original datapoints given
     * @param    final_lpc    array of final linear prediction coefficients
     * 
     */
    public void estimate( Vector<MyPoint> iset, Vector<MyPoint> y_estimate, 
			  double avg, double[] final_lpc)
    {
	   y_estimate.removeAllElements();
	   double amplitude[];
	
	   for(int i = 0; i < iset.size(); i++)
	       y_estimate.addElement(new MyPoint((MyPoint)iset.elementAt(i))); 

	   amplitude = new double[y_estimate.size()];
	   
	   for (int i = 0; i < y_estimate.size(); i++)
	       amplitude[i] = ((MyPoint)y_estimate.elementAt(i)).y;
	   
	   ((MyPoint)y_estimate.firstElement()).y = 0 ;
	   
	   for ( int i = 1; i < y_estimate.size(); i++ )
	   {    
	       int z = i;
  	       double sum  = 0;
	       
	       for( int j = 1; (j <= lporder) && (z > 0); j++, z-- )
		   sum = sum - amplitude[z - 1] * final_lpc[j];
	       
	       ((MyPoint)y_estimate.elementAt(i)).y = sum + avg;
	    }
    }

    /**
     *
     * Compute the actual error from the given data points and the estimated
     * values.
     *
     * @param  y_estimate datapoints of the estimated datapoints
     * @param  iset       original datapoints 
     *
     * @return actual error energy in a double
     */
    public double actual_error (Vector y_estimate, Vector iset)
    {
	double error_value;
	double act_error = (double) 0;
	int j = 0;
	int i = 0;

	// first point and the last point have their x-coordinates same
	// So the difference in y-values is the error value
	//
	error_value = ((MyPoint)y_estimate.firstElement()).y 
	              - ((MyPoint)iset.firstElement()).y;
	act_error = act_error + error_value * error_value; 
	error_value = ((MyPoint)y_estimate.lastElement()).y 
	              - ((MyPoint)iset.lastElement()).y;
	act_error = act_error + error_value * error_value;

	// for next values
	// need to continue till all the user defined points are exhausted
	//
	j++;

	for (i = 1; ( (i < y_estimate.size()) && (j < iset.size()) ); i++)
	{ 
	    while ( (((MyPoint)y_estimate.elementAt(i)).x < 
		     ((MyPoint)iset.elementAt(j)).x) 
		    && ((i < y_estimate.size() - 1) && (j < iset.size())) )
	    {
		    i++;
	    }

	    if ( j < iset.size())
	    {
		if ( ((MyPoint)y_estimate.elementAt(i)).x 
		     == ((MyPoint)iset.elementAt(j)).x ) 
		{
		    error_value = ((MyPoint)y_estimate.elementAt(i)).y 
			          - ((MyPoint)iset.elementAt(j)).y;
		    act_error = act_error + error_value * error_value;
		}

		if ( ((MyPoint)y_estimate.elementAt(i)).x 
		     > ((MyPoint)iset.elementAt(j)).x )
		{
		    double y1 = ((MyPoint)y_estimate.elementAt(i)).y;
		    double y2 = ((MyPoint)y_estimate.elementAt(i - 1)).y;
		    double x1 = ((MyPoint)y_estimate.elementAt(i)).x;
		    double x2 = ((MyPoint)y_estimate.elementAt(i - 1)).x;
		    double x_unknown = ((MyPoint)iset.elementAt(j)).x;
		    error_value = y2 - ( (x2 - x_unknown) * 
					 ( y2 - y1) / (x2 - x1));
		    act_error = act_error + error_value * error_value ;
		}
	    }
	    j++;
	}
	return act_error;
    }
	
    /**
     *
     * Displays LP order, Error Energy and Reflection Coefficients 
     *
     */
    public void step2_display()
    {
	// display the LP order
	//
	pro_box_d.appendMessages("         LP order =  " + lporder + "\n");

	if (set1_d.size() > 0 )
        {
	    int num_pts =  iset1.size();
	    int n = 0;
	    display_result(auto_co_1, ref_coeff_1, final_lpc_1, 
			   actual_err_1, estimate_err_1, n, num_pts);
       	}

	if (set2_d.size() > 0 )
        {
	    int num_pts =  iset2.size();	        
	    int n = 1;
	    display_result(auto_co_2, ref_coeff_2, final_lpc_2, 
			   actual_err_2, estimate_err_2, n, num_pts);
	}

	if (set3_d.size() > 0 )
        {
	    int num_pts =  iset3.size();
	    int n = 2;
	    display_result(auto_co_3, ref_coeff_3, final_lpc_3, 
			   actual_err_3, estimate_err_3, n, num_pts);
	}

	if (set4_d.size() > 0 )
        {
	    int num_pts =  iset4.size();
	    int n = 3;
	    display_result(auto_co_4, ref_coeff_4, final_lpc_4,
			   actual_err_4, estimate_err_4, n, num_pts);
	}
    }

    /**
     * Display the results in the process box
     *
     * @param    auto_coeff Auto Correlation Coefficients
     * @param    refCoef Refelction Coefficient 
     * @param    final_lpc Linear Prediction Coefficients
     * @param    est_err Estimated Error
     * @param    act_err Actual Error
     * @param    length Length of the data points
     *
     */ 
    public void display_result( double[] auto_coeff, double[] refCoef, 
				double[] final_lpc, double est_err, 
				double act_err, int index, int length)
    { 
	pro_box_d.appendMessages("\n" + "         Class  " + index + " : \n");
	pro_box_d.appendMessages("         Number of Points = " + length 
				 + "\n");

	pro_box_d.appendMessages( "         AutoCorrelation Coefficients:" 
				  + "\n");
	pro_box_d.appendMessages("              [ ");
	for (int i = 0 ; i <= lporder; i++)
	    if (i == lporder)
		pro_box_d.appendMessages(MathUtil.SetDecimal(auto_coeff[i], 3)
					 + " ");
	    else
		pro_box_d.appendMessages(MathUtil.SetDecimal(auto_coeff[i], 3)
					 + ", ");
	pro_box_d.appendMessages(" ]");

	pro_box_d.appendMessages("\n" + "         Reflection Coefficients:" 
				 + "\n");
	pro_box_d.appendMessages("              [ ");
	for (int i = 1 ; i <= lporder; i++)
	    if (i == lporder)
		pro_box_d.appendMessages(MathUtil.SetDecimal(refCoef[i], 3) 
					 + " ");
	    else
		pro_box_d.appendMessages(MathUtil.SetDecimal(refCoef[i], 3) 
					 + ", ");
	pro_box_d.appendMessages(" ]");	

	pro_box_d.appendMessages("\n" + "         Prediction Coefficients:" 
				 + "\n");
	pro_box_d.appendMessages("              [ ");
	for (int i = 0 ; i <= lporder; i++)
	    if (i == lporder )
	    pro_box_d.appendMessages(MathUtil.SetDecimal(final_lpc[i], 3) 
				     + " ");
	    else
		pro_box_d.appendMessages(MathUtil.SetDecimal(final_lpc[i], 3) 
					 + ", ");
	pro_box_d.appendMessages(" ]");

	pro_box_d.appendMessages("\n" + "         Estimated Error Energy  = " 
				 + MathUtil.SetDecimal(act_err, 3)); 
	pro_box_d.appendMessages("\n" 
				 + "         Actual Error Energy         = " 
				 + MathUtil.SetDecimal(est_err, 3) + "\n"); 

    }

    /**
     *
     * displays the predicted signal
     *
     * @return   True
     * 
     */
    boolean step3()
    {
	// The display needs to be changed to draw a line between
	// the two points on the waveform.
	// The method is:
	// 1. Get the first two estimated points.
	// 2. Draw a straight line between these two points.
	// 3. Drop the first point, now take the next point.
	// 4. Continue with step 2 and 3 in the same way, 
	//    till you come to the end.
	//
	if(set1_d.size() > 0)
	    output_panel_d.addOutput( y_estimate1, Classify.PTYPE_LINE, 
				      Color.black);
	if(set2_d.size() > 0)
	    output_panel_d.addOutput( y_estimate2, Classify.PTYPE_LINE, 
				      Color.pink);
	if(set3_d.size() > 0)
	    output_panel_d.addOutput( y_estimate3, Classify.PTYPE_LINE, 
				      Color.cyan);
	if(set4_d.size() > 0)
	    output_panel_d.addOutput( y_estimate4, Classify.PTYPE_LINE, 
				      Color.magenta);

	output_panel_d.repaint();

	// displaying "Algorithm Complete" 
	//
	pro_box_d.appendMessages("Algorithm Complete " + "\n");

       	return true;
    }
}
