/*
 * AlgorithmPCA2.java last edited Ryan Irwin V6.0
 * @(#) AlgorithmPCA2.java  1.10 03/15/03
 *
 */


// import java packages
//
import java.awt.*;
import java.util.*;

/**
 * Operation of the PCA Class-Independent Algorithm
 */
public class AlgorithmPCA2 extends Algorithm
{
    //-----------------------------------------------------------------
    //
    // instance data members
    //
    //-----------------------------------------------------------------

    Matrix PCA1_d;
    Matrix PCA2_d;
    Matrix PCA3_d;
    Matrix PCA4_d;

    Matrix CPCA1_d;
    Matrix CPCA2_d;
    Matrix CPCA3_d;
    Matrix CPCA4_d;

    // vector of support region points
    //
    Vector<MyPoint> support_vectors_d = new Vector<MyPoint>();
    Vector<MyPoint> decision_regions_d = new Vector<MyPoint>();
    int output_canvas_d[][];
    String algo_id = "AlgorithmPCA";

    //------------------------------------------------------------------
    //
    // public class methods
    //
    //------------------------------------------------------------------

    /**
     * Overrides the initialize() method in the base class.  Initializes
     * member data and prepares for execution of first step.  This method
     * "resets" the algorithm.
     *
     * @return true
     */
    public boolean initialize()
    {
	// Debug
	//
	// System.out.println("AlgorithmPCA2 : initialize()");

	support_vectors_d = new Vector<MyPoint>();
	point_means_d = new Vector<MyPoint>();
	decision_regions_d = new Vector<MyPoint>();
	step_count = 3;


	// check the data points
	//
	if (output_panel_d == null)
	{
	    return false;
	}
	
	// add the process description for the PCA2 algorithm
	//
	if (description_d.size() == 0)
      	{
	    String str = new String("   0. Initialize the original data.");
	    description_d.addElement(str);
	    
	    str = new String("   1. Displaying the original data.");
	    description_d.addElement(str);
	    
	    str = new String("   2. Computing the means and support regions.");
	    description_d.addElement(str);
	    
	    str = new String("   3. Computing the decision regions.");
	    description_d.addElement(str);
	}
	
	// append message to process box
	//
	pro_box_d.appendMessage("Class Dependent PCA :" + "\n");

	PCA1_d = null;
	PCA2_d = null;
	PCA3_d = null;
	PCA4_d = null;

	CPCA1_d = null;
	CPCA2_d = null;
	CPCA3_d = null;
	CPCA4_d = null;

	// set the data points for this algorithm
	//
	//	set1_d = (Vector)data_points_d.dset1.clone();
	//	set2_d = (Vector)data_points_d.dset2.clone();
	//	set3_d = (Vector)data_points_d.dset3.clone();
	//	set4_d = (Vector)data_points_d.dset4.clone();
	//
	set1_d = data_points_d.dset1;
	set2_d = data_points_d.dset2;
	set3_d = data_points_d.dset3;
	set4_d = data_points_d.dset4;

	// reset values
	//
	support_vectors_d = new Vector<MyPoint>();
	decision_regions_d = new Vector<MyPoint>();

	// set the step index
	//
	step_index_d = 0;

	// append message to process box
	//
	pro_box_d.appendMessage((String)description_d.get(step_index_d));

	// exit gracefully
	//
	return true;
    }

  
    /**
     * Implementation of the run function from the Runnable interface.
     * Determines what the current step is and calls the appropriate method.
     */
    public void run()
    {
	// Debug
	//
	// System.out.println(algo_id + ": run()");
	    
	if (step_index_d == 1)
	{
	    disableControl();
	    step1();
	    enableControl();
	}
	
	else if (step_index_d == 2)
        {
	    disableControl();
	    step2(); 
	    enableControl();
	}
	
        else if (step_index_d == 3)
	{
	    disableControl();
	    step3();
	    pro_box_d.appendMessage("   Algorithm Complete");
	    enableControl(); 
	}
	
	// exit gracefully
	//
	return;
    }
    
    /**
     *
     * step one of the algorithm
     *
     * @return true
     */
    boolean step1()
    {

	// Debug
	//
	// System.out.println(algo_id + ": step1()");
	    
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(1);
	pro_box_d.setProgressCurr(0);

	scaleToFitData();

	// Display original data
	//
	output_panel_d.addOutput(set1_d, Classify.PTYPE_INPUT, 
				 data_points_d.color_dset1);
	output_panel_d.addOutput(set2_d, Classify.PTYPE_INPUT,
				 data_points_d.color_dset2);
	output_panel_d.addOutput(set3_d, Classify.PTYPE_INPUT,
				 data_points_d.color_dset3);
	output_panel_d.addOutput(set4_d, Classify.PTYPE_INPUT, 
				 data_points_d.color_dset4);
	    
	// step 1 completed
	//
	pro_box_d.setProgressCurr(1);
	output_panel_d.repaint();
	    
	// exit gracefully
	//
	return true;
    }

    /**
     *
     * step two of the algorithm
     *
     * @return true
     *
     */
    boolean step2()
    {
	// Debug
	//
	// System.out.println(algo_id + " : step2()");

	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(20);
	pro_box_d.setProgressCurr(0);
		
	transformPCA2();
	printMatrices();
	computeMeans();
		

	// display means
	//
	output_panel_d.addOutput(point_means_d, Classify.PTYPE_OUTPUT_LARGE, 
				 Color.black);
		
	// display support vectors
	//
	output_panel_d.addOutput(support_vectors_d, Classify.PTYPE_INPUT, 
				 Color.cyan);
	
	pro_box_d.setProgressCurr(20);
	output_panel_d.repaint();
	
	return true;
    }

    /**
     *
     * step three of the algorithm
     *
     * @return true
     */
    boolean step3()
    {
	// Debug
	//
	// System.out.println(algo_id + " : step3()");

	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(20);
	pro_box_d.setProgressCurr(0);
	    
	// compute the decision regisions
	//
	computeDecisionRegions();
	    
	// compute errors
	//
	computeErrors();
	    
	// display support vectors
	//
	output_panel_d.addOutput(decision_regions_d, Classify.PTYPE_INPUT, 
				 new Color(255, 200, 0));
	    
	output_panel_d.repaint();

	return true;
    }

   /**
    *
    * Transforms a given set of points to a new space
    * using the class dependent principal component analysis algorithm
    *
    */
    public void transformPCA2()
    {
	// Debug
	//
	// System.out.println(algo_id + " : transformPCA2()");
	    
	// declare local variables
	//
	int size = 0;
	int xsize1 = 0;
	int ysize1 = 0;
	int xsize2 = 0;
	int ysize2 = 0;
	int xsize3 = 0;
	int ysize3 = 0;
	int xsize4 = 0;
	int ysize4 = 0;

	double xval1 = 0.0;
	double yval1 = 0.0;
	double xval2 = 0.0;
	double yval2 = 0.0;
	double xval3 = 0.0;
	double yval3 = 0.0;
	double xval4 = 0.0;
	double yval4 = 0.0;

	double xmean1 = 0.0;
	double ymean1 = 0.0;
	double xmean2 = 0.0;
	double ymean2 = 0.0;
	double xmean3 = 0.0;
	double ymean3 = 0.0;
	double xmean4 = 0.0;
	double ymean4 = 0.0;

	double xval = 0.0;
	double yval = 0.0;

	// declare the covariance object
	//
	Covariance cov = new Covariance();

	// declare an eigen object
	//
	// Since Eigen is a class of static member functions 
	// it is not correct to instantiate it - Phil T. 6-23-03 
	// Eigen eigen = new Eigen();
	
	// declare arrays for the eigenvalues
	//
	double eigVal1[] = null;
	double eigVal2[] = null;
	double eigVal3[] = null;
	double eigVal4[] = null;

	// declare an array to store the eigen vectors
	//
	double eigVec[] = new double[2];

	// declare arrays to store the samples
	//
	double x[] = null;
	double y[] = null;

	// get the samples from the first data set
	//
	size = set1_d.size();

	// increment the variable count for the first data set
	//
	xsize1 += size;
	ysize1 += size;

	// initialize arrays to store the samples
	//
	x = new double[size];
	y = new double[size];

	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
	{
		
	    MyPoint p = (MyPoint)set1_d.elementAt(i);
	    xval1 += p.x;
	    yval1 += p.y;
	    x[i] = p.x;
	    y[i] = p.y;
		
	}
	    
	if (size > 0)
	{
		
	    // declare the covariance matrix
	    //
	    Matrix covariance = new Matrix();
	    covariance.row = covariance.col = 2;
	    covariance.Elem = new double[2][2];
		
	    // declare matrix objects
	    //
	    Matrix T = new Matrix();
	    Matrix M = new Matrix();
	    Matrix W = new Matrix();
		
	    // allocate memory for the matrix elements
	    //
	    T.Elem = new double[2][2];
	    M.Elem = new double[2][2];
	    W.Elem = new double[2][2];
		
	    // initialize the transformation matrix dimensions
	    //
	    W.row = 2;
	    W.col = 2;
		
	    // reset the matrices
	    //
	    W.resetMatrix();
		
	    // compute the covariance matrix of the first data set
	    //
	    covariance.Elem = cov.computeCovariance(x, y);
	    CPCA1_d = covariance;
			
	    // initialize the matrix needed to compute the eigenvalues
	    //
	    T.initMatrix(covariance.Elem, 2, 2);

	    // make a copy of the original matrix
	    //
	    M.copyMatrix(T);
		
	    // compute the eigen values
	    //
	    // Changed eigen to Eigen 
	    // since member function is static - Phil T. 6-23-03
	    eigVal1 = Eigen.compEigenVal(T);

	    // compute the eigen vectors
	    //
	    for (int i = 0; i < 2; i++)
	    {
		Eigen.calcEigVec(M, eigVal1[i], eigVec);
		for (int j = 0; j < 2; j++)
		{
		    W.Elem[j][i] = eigVec[j] / Math.sqrt(eigVal1[i]);
		}
	    }
		
	    // save the transformation matrix 
	    //
	    PCA1_d = W;
	}
	    
	// get the samples from the first data set
	//
	size = set2_d.size();
	    
	// increment the variable count for the second data set
	//
	xsize2 += size;
	ysize2 += size;
	    
	// initialize arrays to store the samples
	//
	x = new double[size];
	y = new double[size];
	
	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
	{
	    MyPoint p = (MyPoint)set2_d.elementAt(i);
	    xval2 += p.x;
	    yval2 += p.y;
	    x[i] = p.x;
	    y[i] = p.y;
	}
	    
	if (size > 0)
	{
		
	    // declare the covariance matrix
	    //
	    Matrix covariance = new Matrix();
	    covariance.row = covariance.col = 2;
	    covariance.Elem = new double[2][2];
		
	    // declare matrix objects
	    //
	    Matrix T = new Matrix();
	    Matrix M = new Matrix();
	    Matrix W = new Matrix();
		
	    // allocate memory for the matrix elements
	    //
	    T.Elem = new double[2][2];
	    M.Elem = new double[2][2];
	    W.Elem = new double[2][2];
			
	    // initialize the transformation matrix dimensions
	    //
	    W.row = 2;
	    W.col = 2;
		
	    // reset the matrices
	    //
	    W.resetMatrix();
		
	    // compute the covariance matrix of the first data set
	    //
	    covariance.Elem = cov.computeCovariance(x, y);
	    CPCA2_d = covariance;
		
	    // initialize the matrix needed to compute the eigenvalues
	    //
	    T.initMatrix(covariance.Elem, 2, 2);
		
	    // make a copy of the original matrix
	    //
	    M.copyMatrix(T);
		
	    // compute the eigen values
	    //
	    // Changed eigen to Eigen since member 
	    // function is static - Phil T. 6-23-03
	    eigVal2 = Eigen.compEigenVal(T);
		
	    // compute the eigen vectors
	    //
	    for (int i = 0; i < 2; i++)
	    {
		// Changed eigen to Eigen 
		// since member function is static - Phil T. 6-23-03	
		Eigen.calcEigVec(M, eigVal2[i], eigVec);
		for (int j = 0; j < 2; j++)
		{
		    W.Elem[j][i] = eigVec[j] / Math.sqrt(eigVal2[i]);
		}
	    }
	    
	    // save the transformation matrix 
	    //
	    PCA2_d = W;
	}
	    
	// get the samples from the first data set
	//
	size = set3_d.size();

	// increment the variable count for the third data set
	//
	xsize3 += size;
	ysize3 += size;
	
	// initialize arrays to store the samples
	//
	x = new double[size];
	y = new double[size];

	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
        {
	    MyPoint p = (MyPoint)set3_d.elementAt(i);
	    xval3 += p.x;
	    yval3 += p.y;
	    x[i] = p.x;
	    y[i] = p.y;
	}

	if (size > 0)
	{

	    // declare the covariance matrix
	    //
	    Matrix covariance = new Matrix();
	    covariance.row = covariance.col = 2;
	    covariance.Elem = new double[2][2];

	    // declare matrix objects
	    //
	    Matrix T = new Matrix();
	    Matrix M = new Matrix();
	    Matrix W = new Matrix();
	    
	    // allocate memory for the matrix elements
	    //
	    T.Elem = new double[2][2];
	    M.Elem = new double[2][2];
	    W.Elem = new double[2][2];

	    // initialize the transformation matrix dimensions
	    //
	    W.row = 2;
	    W.col = 2;

	    // reset the matrices
	    //
	    W.resetMatrix();

	    // compute the covariance matrix of the first data set
	    //
	    covariance.Elem = cov.computeCovariance(x, y);
	    CPCA3_d = covariance;
	    
	    // initialize the matrix needed to compute the eigenvalues
	    //
	    T.initMatrix(covariance.Elem, 2, 2);
	    
	    // make a copy of the original matrix
	    //
	    M.copyMatrix(T);
	    
	    // compute the eigen values
	    //
	    // Changed eigen to Eigen
	    // since member function is static - Phil T. 6-23-03
	    //
	    eigVal3 = Eigen.compEigenVal(T);
	    
	    // compute the eigen vectors
	    //
	    for (int i = 0; i < 2; i++)
	    {
		// Changed eigen to Eigen 
		// since member function is static - Phil T. 6-23-03
		//
		Eigen.calcEigVec(M, eigVal3[i], eigVec);
		for (int j = 0; j < 2; j++)
		{
		    W.Elem[j][i] = eigVec[j] / Math.sqrt(eigVal3[i]);
		}
	    }
	    
	    // save the transformation matrix 
	    //
	    PCA3_d = W;
	}
	    
	// get the samples from the first data set
	//
	size = set4_d.size();
	    
	// increment the variable count for the forth data set
	//
	xsize4 += size;
	ysize4 += size;
	    
	// initialize arrays to store the samples
	//
	x = new double[size];
	y = new double[size];
	    
	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
	{
	    MyPoint p = (MyPoint)set4_d.elementAt(i);
	    xval4 += p.x;
	    yval4 += p.y;
	    x[i] = p.x;
	    y[i] = p.y;
	}
	    
	if (size > 0)
	{
		
	    // declare the covariance matrix
	    //
	    Matrix covariance = new Matrix();
	    covariance.row = covariance.col = 2;
	    covariance.Elem = new double[2][2];
		
	    // declare matrix objects
	    //
	    Matrix T = new Matrix();
	    Matrix M = new Matrix();
	    Matrix W = new Matrix();
		
	    // allocate memory for the matrix elements
	    //
	    T.Elem = new double[2][2];
	    M.Elem = new double[2][2];
	    W.Elem = new double[2][2];
		
	    // initialize the transformation matrix dimensions
	    //
	    W.row = 2;
	    W.col = 2;
	    
	    // reset the matrices
	    //
	    W.resetMatrix();
	    
	    // compute the covariance matrix of the first data set
	    //
	    covariance.Elem = cov.computeCovariance(x, y);
	    CPCA4_d = covariance;
		
	    // initialize the matrix needed to compute the eigenvalues
	    //
	    T.initMatrix(covariance.Elem, 2, 2);
		
	    // make a copy of the original matrix
	    //
	    M.copyMatrix(T);

	    // compute the eigen values
	    //
	    // Changed eigen to Eigen 
	    // since member function is static - Phil T. 6-23-03
	    eigVal4 = Eigen.compEigenVal(T);
		
	    // compute the eigen vectors
	    //
	    for (int i = 0; i < 2; i++)
	    {
		// Changed eigen to Eigen 
		// since member function is static - Phil T. 6-23-03
		//
		Eigen.calcEigVec(M, eigVal4[i], eigVec);
		for (int j = 0; j < 2; j++)
		{
		    W.Elem[j][i] = eigVec[j] / Math.sqrt(eigVal4[i]);
		}
	    }
		
	    // save the transformation matrix 
	    //
	    PCA4_d = W;
	}
	    
	// determine the local mean of each data set
	//
	if (xsize1 > 0 && ysize1 > 0)
	{
	    xmean1 = (double) (xval1 / xsize1);
	    ymean1 = (double) (yval1 / ysize1);
	}

	if (xsize2 > 0 && ysize2 > 0)
	{
	    xmean2 = (double) (xval2 / xsize2);
	    ymean2 = (double) (yval2 / ysize2);
	}

	if (xsize3 > 0 && ysize3 > 0)
	{
	    xmean3 = (double) (xval3 / xsize3);
	    ymean3 = (double) (yval3 / ysize3);
	}

	if (xsize4 > 0 && ysize4 > 0)
	{
	    xmean4 = (double) (xval4 / xsize4);
	    ymean4 = (double) (yval4 / ysize4);
	}
	
	// determine the support regions data points
	//
	double val[][] = new double[2][1];
	
	Matrix invT = new Matrix();
	Matrix supp = new Matrix();
	Matrix temp = new Matrix();
	
	// determine if the second transformation matrix is computed
	//
	if (PCA1_d != null)
	{
	    
	    // set up the angle with which to rotate the axis
	    //
	    double theta = 0.0;
	    double alpha = CPCA1_d.Elem[0][0] - CPCA1_d.Elem[1][1];
	    double beta = -2 * CPCA1_d.Elem[0][1];
	    
	    if (eigVal1[0] > eigVal1[1])
	    {
		theta = Math.atan2((alpha - Math.sqrt((alpha * alpha) 
						      + (beta * beta))), beta);
	    }
	    else
	    {
		theta = Math.atan2((alpha + Math.sqrt((alpha * alpha) 
						      + (beta * beta))), beta);
	    }

	    // compute the inverse of the transformation matrix
	    //
	    PCA1_d.invertMatrix(invT);
	    
	    // loop through all points on the circumference of the 
	    // gaussian sphere in the transformed space
	    //
	    for (int i = 0; i < 360; i++)
	    {
		// get the x and y co-ordinates
		//
		val[0][0] = 1.5 * Math.cos(i);
		val[1][0] = 1.5 * Math.sin(i);
		
		// set up the points as a matrix in order for multiplication
		//
		temp.initMatrix(val, 2, 1);
		
		// transform the points from the feature space back to the
		// original space to create the support region for the data set
		//
		invT.multMatrix(temp, supp);

		// rotate the points after transforming them to the new space
		//
		xval = (supp.Elem[0][0] * Math.cos(theta)) 
		     - (supp.Elem[1][0] * Math.sin(theta));
		yval = (supp.Elem[0][0] * Math.sin(theta)) 
		     + (supp.Elem[1][0] * Math.cos(theta));
		
		// shift the points to the class mean
		//
		xval = xval + xmean1;
		yval = yval + ymean1;
		    
		// add the point to the support region vector
		//
		MyPoint pt = new MyPoint(xval, yval);
		support_vectors_d.addElement(pt);
	    }
	}
	
	// determine if the second transformation matrix is computed
	//
	if (PCA2_d != null)
        {
	    
	    // set up the angle with which to rotate the axis
	    //
	    double theta = 0.0;
	    double alpha = CPCA2_d.Elem[0][0] - CPCA2_d.Elem[1][1];
	    double beta = -2 * CPCA2_d.Elem[0][1];
	    
	    if (eigVal2[0] > eigVal2[1])
	    {
		theta = Math.atan2((alpha - Math.sqrt((alpha * alpha) 
						      + (beta * beta))), beta);
	    }
	    else
	    {
		theta = Math.atan2((alpha + Math.sqrt((alpha * alpha) 
						      + (beta * beta))), beta);
	    }
	    
	    // compute the inverse of the transformation matrix
	    //
	    PCA2_d.invertMatrix(invT);
	    
	    // loop through all points on the circumference of the 
	    // gaussian sphere in the transformed space
	    //
	    for (int i = 0; i < 360; i++)
	    {
		
		// get the x and y co-ordinates
		//
		val[0][0] = 1.5 * Math.cos(i);
		val[1][0] = 1.5 * Math.sin(i);
		
		// set up the points as a matrix in order for multiplication
		//
		temp.initMatrix(val, 2, 1);
		
		// transform the points from the feature space back to the
		// original space to create the support region for the data set
		//
		invT.multMatrix(temp, supp);
		
		// rotate the points after transforming them to the new space
		//
		xval = (supp.Elem[0][0] * Math.cos(theta)) 
		     - (supp.Elem[1][0] * Math.sin(theta));
		yval = (supp.Elem[0][0] * Math.sin(theta)) 
		     + (supp.Elem[1][0] * Math.cos(theta));
		
		// shift the points to the class mean
		//
		xval = xval + xmean2;
		yval = yval + ymean2;
		
		// add the point to the support region vector
		//
		MyPoint pt = new MyPoint(xval, yval);
		support_vectors_d.addElement(pt);
	    }
	}
	
	// determine if the third transformation matrix is computed
	//
	if (PCA3_d != null)
	{
	    
	    // set up the angle with which to rotate the axis
	    //
	    double theta = 0.0;
	    double alpha = CPCA3_d.Elem[0][0] - CPCA3_d.Elem[1][1];
	    double beta = -2 * CPCA3_d.Elem[0][1];
	    
	    if (eigVal3[0] > eigVal3[1])
	    {
		theta = Math.atan2((alpha - Math.sqrt((alpha * alpha) 
						      + (beta * beta))), beta);
	    }
	    else
	    {
		theta = Math.atan2((alpha + Math.sqrt((alpha * alpha) 
						      + (beta * beta))), beta);
	    }
	    
	    // compute the inverse of the transformation matrix
	    //
	    PCA3_d.invertMatrix(invT);
	    
	    // loop through all points on the circumference of the 
	    // gaussian sphere in the transformed space
	    //
	    for (int i = 0; i < 360; i++)
	    {
		
		// get the x and y co-ordinates
		//
		val[0][0] = 1.5 * Math.cos(i);
		val[1][0] = 1.5 * Math.sin(i);
		
		// set up the points as a matrix in order for multiplication
		//
		temp.initMatrix(val, 2, 1);
		
		// transform the points from the feature space back to the
		// original space to create the support region for the data set
		//
		invT.multMatrix(temp, supp);
		
		// rotate the points after transforming them to the new space
		//
		xval = (supp.Elem[0][0] * Math.cos(theta)) 
		     - (supp.Elem[1][0] * Math.sin(theta));
		yval = (supp.Elem[0][0] * Math.sin(theta)) 
		     + (supp.Elem[1][0] * Math.cos(theta));
		
		// shift the points to the class mean
		//
		xval = xval + xmean3;
		yval = yval + ymean3;
		
		// add the point to the support region vector
	        //
		MyPoint pt = new MyPoint(xval, yval);
		support_vectors_d.addElement(pt);
	    }
	}
	
	// determine if the forth transformation matrix is computed
	//
	if (PCA4_d != null)
	{
	    
	    // set up the angle with which to rotate the axis
	    //
	    double theta = 0.0;
	    double alpha = CPCA4_d.Elem[0][0] - CPCA4_d.Elem[1][1];
	    double beta = -2 * CPCA4_d.Elem[0][1];
	    
	    if (eigVal4[0] > eigVal4[1])
	    {
		theta = Math.atan2((alpha - Math.sqrt((alpha * alpha) 
						      + (beta * beta))), beta);
	    }
	    else
	    {
		theta = Math.atan2((alpha + Math.sqrt((alpha * alpha) 
						      + (beta * beta))), beta);
	    }
	    
	    // compute the inverse of the transformation matrix
	    //
	    PCA4_d.invertMatrix(invT);
	    
	    // loop through all points on the circumference of the 
	    // gaussian sphere in the transformed space
	    //
	    for (int i = 0; i < 360; i++)
	    {
		
		// get the x and y co-ordinates
		//
		val[0][0] = 1.5 * Math.cos(i);
		val[1][0] = 1.5 * Math.sin(i);
		
		// set up the points as a matrix in order for multiplication
		//
		temp.initMatrix(val, 2, 1);
		
		// transform the points from the feature space back to the
		// original space to create the support region for the data set
	        //
		invT.multMatrix(temp, supp);
		
		// rotate the points after transforming them to the new space
	        //
		xval = (supp.Elem[0][0] * Math.cos(theta)) 
		     - (supp.Elem[1][0] * Math.sin(theta));
		yval = (supp.Elem[0][0] * Math.sin(theta)) 
		     + (supp.Elem[1][0] * Math.cos(theta));
		
		// shift the points to the class mean
		//
		xval = xval + xmean4;
		yval = yval + ymean4;
		
		// add the point to the support region vector
		//
		MyPoint pt = new MyPoint(xval, yval);
		support_vectors_d.addElement(pt);
	    }
	}
    }
    
    /**
     * Computes the line of discrimination for the classification 
     * algorithms when the corresponding flags have been initialized
     *
     */
    public void computeDecisionRegions()
    {
	// Debug
	//
	// System.out.println(algo_id + " : computeDecisionRegions()");
	
	DisplayScale scale = output_panel_d.disp_area_d.getDisplayScale();
	
	double currentX = scale.xmin;
	double currentY = scale.ymin;
	
	// set precision
	//
	int outputWidth = output_panel_d.disp_area_d.getXPrecision();
	int outputHeight = output_panel_d.disp_area_d.getYPrecision();
	    
	double incrementY = (scale.ymax-scale.ymin)/outputHeight;
	double incrementX = (scale.xmax-scale.xmin)/outputWidth;
	    
	// declare a 2D array to store the class associations
	//
	output_canvas_d = new int[outputWidth][outputHeight];
	    
	// loop through each and every point on the pixmap and
	// determine which class each pixel is associated with
	//
	MyPoint point;
	double dist = 0.0;
	int associated = 0;
	double smallestSoFar = Double.MAX_VALUE;
	    
	int target = 0;
	boolean set1flag = true;
	boolean set2flag = true;
	boolean set3flag = true;

	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(outputWidth);

	for (int i = 0; i < outputWidth; i++)
	{
		
	    // set current status
	    //		
	    currentX += incrementX;
	    currentY = scale.ymin;
	    pro_box_d.setProgressCurr(i);
		
	    for (int j = 0; j < outputHeight; j++)
	    {
		    
		// declare the current pixel point
		//	    
		currentY += incrementY;
		MyPoint pixel = new MyPoint(currentX, currentY);
		smallestSoFar = Double.MAX_VALUE;
		    
		// convert the pixel to the time domain
		//
		double X[][] = new double[1][2];
		X[0][0] = pixel.x;
		X[0][1] = pixel.y;
		    
		// reset the boolean flags
		//
		set1flag = true;
		set2flag = true;
		set3flag = true;
		    
		// find the closest point from the first class
		//
		for (int k = 0; k < point_means_d.size(); k++)
		{
		    
		    // get the first mean point
		    //
		    point = (MyPoint)point_means_d.elementAt(k);

		    // convert the mean point to the time domain
		    //
		    double Y[][] = new double[1][2];
		    Y[0][0] = point.x;
		    Y[0][1] = point.y;

		    // represent the pixel as a matrix
		    //
		    Matrix A = new Matrix();
		    A.initMatrix(X, 1, 2);
			
		    // represent the mean point as a matrix
		    //
		    Matrix B = new Matrix();
		    B.initMatrix(Y, 1, 2);
			
		    // transform the pixel and mean point to the 
		    // feature space
		    //
		    Matrix C = new Matrix();
		    Matrix D = new Matrix();
			
		    // use the transformation matrix of the
		    // first data set
		    //
		    if (set1_d.size() > 0 && set1flag)
		    {
			set1flag = false;
			target = 0;
			A.multMatrix(PCA1_d, C);
			B.multMatrix(PCA1_d, D);
		    }
		    
		    // use the transformation matrix of the
		    // second data set
		    //
		    else if (set2_d.size() > 0 && set2flag)
		    {
			set2flag = false;
			target = 1;
			A.multMatrix(PCA2_d, C);
			B.multMatrix(PCA2_d, D);
		    }
		    
		    // use the transformation matrix of the
		    // third data set
		    //
		    else if (set3_d.size() > 0 && set3flag)
		    {
			set3flag = false;
			target = 2;
			A.multMatrix(PCA3_d, C);
			B.multMatrix(PCA3_d, D);
		    }
		    
		    // use the transformation matrix of the
		    // forth data set
		    //
		    else
		    {
			target = 3;
			A.multMatrix(PCA4_d, C);
			B.multMatrix(PCA4_d, D);
		    }
			
		    // find the distance between the pixel and
		    // mean point
		    //
		    dist = MathUtil.distance(C.Elem[0][0], C.Elem[0][1], 
					     D.Elem[0][0], D.Elem[0][1]);
		    
		    if (dist < smallestSoFar)
		    {
			associated = target;
			smallestSoFar = dist;
		    }
		}
		
		// put and entry in the output canvas array to
		// indicate which class the current pixel is
		// closest to
		//
		output_canvas_d[i][j] = associated;
		
		// add a point to the vector of decision
		// region points if the class that the current
		// point is associated with is different for
		// the class what the previous point was
		// associated with i.e., a transition point
		//
		if (j > 0 && i > 0)
	        {
		    if (associated != output_canvas_d[i][j - 1] 
			|| associated != output_canvas_d[i - 1][j])
		    {
			decision_regions_d.add(pixel);
		    }
		}
	    }
	} // end of the loop
	
    }
    
    /**
     * 
     * Computes the classification error for the data points given
     *
     */
    public void computeErrors()
    {
	
	// declare local variables
	//
	String text;
	double error;
	int samples = 0;
	int samples1 = 0;
	int samples2 = 0;
	int samples3 = 0;
	int samples4 = 0;
	
	int incorrect = 0;
	int incorrect1 = 0;
	int incorrect2 = 0;
	int incorrect3 = 0;
	int incorrect4 = 0;
	
	DisplayScale scale = output_panel_d.disp_area_d.getDisplayScale();
	
	// set scales
	int outputWidth = output_panel_d.disp_area_d.getXPrecision();
	int outputHeight = output_panel_d.disp_area_d.getYPrecision();
	    
	double incrementY = (scale.ymax-scale.ymin)/outputHeight;
	double incrementX = (scale.xmax-scale.xmin)/outputWidth;

	// compute the classification error for the first set
	//
	for (int i = 0; i < set1_d.size(); i++)
	{
	    
	    MyPoint point = (MyPoint)set1_d.elementAt(i);
	    samples1++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 0)
		{
		    incorrect1++;
		}
	    }
	}
	
	if (set1_d.size() > 0)
        {
	    
	    error = ((double)incorrect1 / (double)samples1) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 0:\n"
			   + "          Total number of samples: "
			   + samples1
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect1
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the classification error for the second set
	//
	for (int i = 0; i < set2_d.size(); i++)
        {
	    
	    MyPoint point = (MyPoint)set2_d.elementAt(i);
	    samples2++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 1)
		{
		    incorrect2++;
		}
	    }
	}
	
	if (set2_d.size() > 0)
	{
	    
	    error = ((double)incorrect2 / (double)samples2) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 1:\n"
			   + "          Total number of samples: "
			   + samples2
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect2
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the classification error for the third set
	//
	for (int i = 0; i < set3_d.size(); i++)
        {
	    
	    MyPoint point = (MyPoint)set3_d.elementAt(i);
	    samples3++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 2)
		{
		    incorrect3++;
		}
	    }
	}
	
	if (set3_d.size() > 0)
	{
		
	    error = ((double)incorrect3 / (double)samples3) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 2:\n"
			   + "          Total number of samples: "
			   + samples3
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect3
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the classification error for the forth set
	//
	for (int i = 0; i < set4_d.size(); i++)
        {
	    
	    MyPoint point = (MyPoint)set4_d.elementAt(i);
	    samples4++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 3)
	        {
		    incorrect4++;
		}
	    }
	}
	
	if (set4_d.size() > 0)
	{
	    
	    error = ((double)incorrect4 / (double)samples4) * 100.0;
		
	    text =
		new String(
			   "       Results for class 3:\n"
			   + "          Total number of samples: "
			   + samples4
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect4
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the overall classification error
	//
	samples = samples1 + samples2 + samples3 + samples4;
	incorrect = incorrect1 + incorrect2 + incorrect3 + incorrect4;
	error = ((double)incorrect / (double)samples) * 100.0;
	
	text =
	    new String(
		       "       Overall results:\n"
		       + "          Total number of samples: "
		       + samples
		       + "\n"
		       + "          Misclassified samples: "
		       + incorrect
		       + "\n"
		       + "          Classification error: "
		       + MathUtil.setDecimal(error, 2)
		       + "%");
	
	pro_box_d.appendMessage(text);
	
    }    
    
   /**
    * Appends messages to the pro_box_d variable
    */   
    public void printMatrices()
    {
	double a11, a12, a21, a22;
	String text;
	
	if (set1_d.size() > 0)
	{
	    pro_box_d.appendMessage("      Set 1");
	    
	    a11 = MathUtil.setDecimal(CPCA1_d.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(CPCA1_d.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(CPCA1_d.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(CPCA1_d.Elem[1][1], 2);
	    
	    text = new String("      Covariance matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	    
	    a11 = MathUtil.setDecimal(PCA1_d.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(PCA1_d.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(PCA1_d.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(PCA1_d.Elem[1][1], 2);
	    
	    text = new String("      Transformation matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	}

	if (set2_d.size() > 0)
	{
	    pro_box_d.appendMessage("      Set 2");
	    
	    a11 = MathUtil.setDecimal(CPCA2_d.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(CPCA2_d.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(CPCA2_d.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(CPCA2_d.Elem[1][1], 2);
	    
	    text = new String("      Covariance matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	    
	    a11 = MathUtil.setDecimal(PCA2_d.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(PCA2_d.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(PCA2_d.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(PCA2_d.Elem[1][1], 2);
	    
	    text = new String("      Transformation matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	}

	if (set3_d.size() > 0)
	{
	    pro_box_d.appendMessage("      Set 3");
	    
	    a11 = MathUtil.setDecimal(CPCA3_d.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(CPCA3_d.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(CPCA3_d.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(CPCA3_d.Elem[1][1], 2);
	    
	    text = new String("      Covariance matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	    
	    a11 = MathUtil.setDecimal(PCA3_d.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(PCA3_d.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(PCA3_d.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(PCA3_d.Elem[1][1], 2);
	    
	    text = new String("      Transformation matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	}

	if (set4_d.size() > 0)
	{
	    
	    pro_box_d.appendMessage("      Set 4");
	    
	    a11 = MathUtil.setDecimal(CPCA4_d.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(CPCA4_d.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(CPCA4_d.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(CPCA4_d.Elem[1][1], 2);
	    
	    text = new String("      Covariance matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	    
	    a11 = MathUtil.setDecimal(PCA4_d.Elem[0][0], 2);
	    a12 = MathUtil.setDecimal(PCA4_d.Elem[0][1], 2);
	    a21 = MathUtil.setDecimal(PCA4_d.Elem[1][0], 2);
	    a22 = MathUtil.setDecimal(PCA4_d.Elem[1][1], 2);
	    
	    text = new String("      Transformation matrix:\n" +
			      "         " + a11 + "    " + a12 + "\n" +
			      "         " + a21 + "    " + a22);
	    
	    pro_box_d.appendMessage(text + "\n");
	}
    }
}




