/*
 * @(#) IFAlgorithm.java  6.0 03/10/05
 *
 * Copyright IES,  All Rights Reserved.
 * 
 * This software is the proprietary information of IES  
 * Use of this software for educational and academic purpose 
 * is permitted.
 * 
 * last edited: Ryan Irwin
 *
 */

/**
 * This interface is designed to be the base for all algorithms
 * @version 6.00
 */
public interface IFAlgorithm {

    /**
     * Sets DataPoints given DataPoints type variable. Will be overwritten.
     *
     * @param data_a DataPoints to be set
     */
    boolean setDataPoints(DataPoints data_a);

    /**
     * Initializes member data and prepares for execution of first step.
     * This method "resets" the algorithm. Will be overwritten.
     *
     * @return  true
     */
    boolean initialize();

    /**
     * Directs algorithm to next step. Will be overwritten.
     *
     * @return true
     */
    boolean nextStep();

    /**
     * Directs algorithm to previous step. Will be overwritten.
     *
     * @return true
     */
    boolean prevStep();

    /**
     * Sets the output panel. Will be overwritten.
     *
     * @param out_panel_a OutputPanel to be set
     * @return true
     */
    boolean setOutputPanel( OutputPanel out_panel_a);

    /**
     * Sets the ProcessBox. Will be overwritten.
     *
     * @param pro_box_a ProcessBox to be set
     * @return true
     */
    boolean setProcessBox( ProcessBox pro_box_a);
}
