/**
 * file: OutputPanel.java
 *
 * last edited: Ryan Irwin
 */


// import necessary java libraries
//
import java.awt.*;
import javax.swing.*;
import java.util.*;
import javax.swing.border.*;

//These imports are not needed - Phil T. 6-23-03
//import java.awt.event.*;
//import javax.swing.Timer;

/**
 * OutputPanel acts as the top level container that encompasses the output
 * plot that displays the input classes as well as the decision regions
 *
 * heirarchy: JPanel->SubPanel->OutputPanel
 *
 */
public class OutputPanel extends SubPanel implements Cloneable
{
	static final long serialVersionUID = 0;
    // *********************************************************************
    //
    // declare global variables and components
    //
    // *********************************************************************

    // static members
    //
    public static double Xmax = 2.0; // maximum range of the x axis
    public static double Xmin = -2.0; // minimum range of the x axis
    public static double Ymax = 2.0; // maximum range of the y axis
    public static double Ymin = -2.0; // minimum range of the y axis
    public static double Rx = 0.0; // time per pixel ration of the x axis
    public static double Ry = 0.0; // time per pixel ration of the y axis

    // declare components
    //
    String title; // title for the input panel

    Font currentFont = getFont(); // fonts for the buttons

    Font newFont = new Font(currentFont.getName(), currentFont.getStyle(), 12);

    DisplayArea disp_area_d; // declare a panel for plotting

    static final int GRID_DIV = 20; // division size for the grid

    int width = 0; // width of the canvas
    int height = 0; // height of the canvas

    int zeroX = 0; // zero position on x axis
    int zeroY = 0; // zero position on y axis

    int currObject = 0; // current object that is in focus

    String newline; // newline statement for text area
    String text; // string for process description

    int algorithm = 0; // index to indicate algorithm to be used

    // *********************************************************************
    //
    // declare class constructors
    //
    // *********************************************************************

    /**
     * constructor initializes the samples to be plotted in the signal panel
     *
     */
    OutputPanel()
    {

	super();

	// initialize the panel used for plotting
	//
	disp_area_d = new DisplayArea();

	// set the border for the panel
	//
	title = new String("Output Display:");
	Border titleBorder =
	    BorderFactory.createTitledBorder(
				BorderFactory.createLineBorder(Color.black),
				title,
				TitledBorder.LEFT,
				TitledBorder.ABOVE_TOP,
				newFont,
				Color.black);
	setBorder(titleBorder);
    }

    /**
     * constructor initializes the samples to be plotted in the signal panel
     *
     * @param   toCopy OutputPanel to be initialized
     */
    OutputPanel(OutputPanel toCopy)
    {

	super();

	// initialize the panel used for plotting
	//
	disp_area_d = toCopy.disp_area_d;

	// set the border for the panel
	//
	title = new String("Output Display:");
	Border titleBorder =
	    BorderFactory.createTitledBorder(
				BorderFactory.createLineBorder(Color.black),
				title,
				TitledBorder.LEFT,
				TitledBorder.ABOVE_TOP,
				newFont,
				Color.black);
	setBorder(titleBorder);
    }

    // *********************************************************************
    //
    // declare class methods
    //
    // *********************************************************************

    /**
     * Adds components to the output panel
     *
     */
  public void add_components()
  {

    constrain(
	      this,
	      disp_area_d,
	      0,
	      0,
	      GridBagConstraints.REMAINDER,
	      GridBagConstraints.REMAINDER,
	      GridBagConstraints.BOTH,
	      GridBagConstraints.NORTHWEST,
	      1,
	      1,
	      2,
	      2,
	      2,
	      2);
  }
    
    /**
     * Paints the current data points and if needed the decision regions
     *
     * @param    g graphics object to use
     */
    public void paintComponent(Graphics g)
    {
	super.paintComponent(g);
	
	// g.drawLine(10, 15, 410, 50); draws a line but always
	// near the Text Output Display:. 
	// the line is drawn in the background.
	// idea is to draw a line - refer the AlgorithmLP.java code.
    }
    
    /**
     * Adds points to the output panel
     *
     * @param points_a points to be added to panel
     * @param type_a integer value of type
     * @param color_a color of point
     *
     * @return   true
     *
     */
    public boolean addOutput(Vector<MyPoint> points_a, int type_a, Color color_a)    {

	disp_area_d.output_points_d.addElement(points_a);
	disp_area_d.type_d.addElement(new Integer(type_a));
	disp_area_d.color_d.addElement(color_a);
	
	// init the algorithm
	//
	return true;
    }

    /**
     * Removes last vector of points from output
     */
    public boolean removeOutput() {

	int size = disp_area_d.output_points_d.size();

	disp_area_d.output_points_d.removeElementAt(size - 1);
	disp_area_d.type_d.removeElementAt(size - 1);
	disp_area_d.color_d.removeElementAt(size - 1);

	return true;	
    }

    /**
     * Gets the width on the output canvas
     * method: getDisplayWidth
     *
     * @return  width of the output plot
     */
    public int getDisplayWidth()
    {
	
	// return the width 
	//
	return disp_area_d.getWidth();
    }

    /**
     * Gets the height on the output canvas
     *
     * @return  height of the output plot
     */
    public int getDisplayHeight()
    {

	// return the width 
	//
	return disp_area_d.getHeight();
    }

    /**
     * Gets the X direction Ratio
     *
     * @return  X ratio of the output plot
     */
    public double getXRatio()
    {
	
	Rx = (Xmax - Xmin) / getDisplayWidth();
	
	// return the width 
	//
	return Rx;
  }

    /**
     * Gets the Y direction Ratio
     *
     * @return  Y Ratio of the output plot
     */
    public double getYRatio()
    {
	
	Ry = (Ymax - Ymin) / getDisplayWidth();
	
	// return the width 
	//
	return Ry;
    }

    /**
     * Clears the output panel screen
     *
     */
    public void clear()
    {
	super.repaint();
	disp_area_d.clear();
	disp_area_d.repaint();
    }

    /**
     * Redraws the output panel screen
     */
    public void refresh()
    {
	super.repaint();
	disp_area_d.repaint();
    }

    /**
     * Duplicates the OutputPanel Object
     *
     * @return  clone of OutputPanel
     */
    public OutputPanel copy()
    {
	try 
	{
	    return (OutputPanel)clone();
	}
	catch (CloneNotSupportedException e) 
	    {}

	return null;	  
    }
}
